from typing import Any, Dict, Tuple

import torch.fx as fx
from torch.fx.node import Argument, Target
from torch.fx.proxy import Proxy

from dmx.compressor.numerical import CastTo
from dmx.compressor.sparse import Sparsify
from .utils import *


class InputOutputTransformer(fx.Transformer):
    """
    A transformer that transforms the module by adding additional ops, which includes:
    - casting
    - approximator
    - sparsifier

    Args:
        module (fx.GraphModule): module to be added with additional ops.
        scopeDict (Optional[dict]): Dictionary that maps node name to scope. Defaults to None.
        cfg (Optional[str]): config file for setting the added ops formats. Defaults to None.

    Attributes:
        module (fx.GraphModule): module to be added with additional ops.
        scopeDict (Optional[str]): Dictionary that maps node name to scope.
        cfg (Optional[str]): config file for setting the added ops formats.

    """

    def __init__(self, module: fx.GraphModule, scopeDict: dict = None, cfg=None):
        super().__init__(module)
        self.scopeDict = scopeDict
        self.config = None
        # if cfg:
        #     self.config = DmxConfig().from_yaml(cfg)
        self.module = module

    def placeholder(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Proxy:
        """
        Execute a ``placeholder`` node. Adds input_cast module after the node.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the last node added (input_cast) and the tracer of the new graph
        """
        assert isinstance(target, str)
        placeholder_node = self.new_graph.placeholder(target)
        # Default input cast
        layer = self.scopeDict[placeholder_node.name][0]
        cast_name = target + "_cast"
        cast_format = "SAME"
        # Find input_cast format in cfg if exists
        if self.config:
            layer_key = layer.split("__")[-1]
            if layer_key and layer_key in self.config:
                cast_format = self.config[layer_key]["input_format"]

        self.module.add_submodule(cast_name, CastTo(format=cast_format))
        placeholder_node_cast = self.new_graph.create_node(
            "call_module", cast_name, args=(placeholder_node,)
        )
        return Proxy(placeholder_node_cast, self.tracer)

    def output(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute an ``output`` node. Insert an output_cast module before the node.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns:
            A Proxy containing the last node added (output) and the tracer of the new graph
        """
        output_node = self.new_graph.output(target)
        # Default output cast
        layer = self.scopeDict[output_node.name][0]
        cast_name = target + "_cast"
        cast_format = "SAME"
        # Find output_cast format in cfg if exists
        if self.config:
            layer_key = layer.split("__")[-1]
            if layer_key and layer_key in self.config:
                cast_format = self.config[layer_key]["output_format"]

        self.module.add_submodule(cast_name, CastTo(format=cast_format))
        self.new_graph.inserting_before(output_node)
        output_node_cast = self.new_graph.create_node(
            "call_module", cast_name, args=(output_node.prev,)
        )
        output_node.args = (output_node_cast,)
        self.new_graph.erase_node(output_node)
        return Proxy(output_node_cast, self.tracer)

    def get_attr(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Proxy:
        """
        Execute a ``get_attr`` node. Insert a weight_cast/bias_cast after the node. Insert additional weight_sparsifier node
        if the current node is a weight node and sparsifier format is specified in the cfg file.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns
            A Proxy containing the last node added (output) and the tracer of the new graph
        """
        assert isinstance(target, str)
        get_attr_node = self.new_graph.get_attr(target)
        prev_node = get_attr_node
        # Default cast
        layer = self.scopeDict[get_attr_node.name][0]
        cast_name = target + "_cast"
        cast_format = "SAME"
        # Default sparsifier
        sparsify_format = "DENSE"
        sparsify_name = target + "_sparsifier"
        layer_key = layer.split("__")[-1]
        # Find casting and sparsifier format in config if exists
        if self.config:
            if layer_key and layer_key in self.config:
                if "weight" in target:
                    if "weight_format" in self.config[layer_key]:
                        cast_format = self.config[layer_key]["weight_format"]
                    if "weight_sparseness" in self.config[layer_key]:
                        # Add sparsifier
                        sparsify_format = self.config[layer_key]["weight_sparseness"]
                else:
                    if "bias_format" in self.config[layer_key]:
                        cast_format = self.config[layer_key]["bias_format"]

        self.module.add_submodule(cast_name, CastTo(format=cast_format))

        # Add sparsifier and approximator for weight nodes if needed
        if "weight" in target:
            # Sparsifier submodules needed to be added separately even for default as tensor size
            # is not the same for every layer
            if layer.split("__")[-1] == "model":
                tensor_size = self.module.get_parameter(target).size()
            else:
                tensor_size = (
                    self.module.get_submodule(layer.split("__")[-1])
                    .get_parameter(target.split(".")[-1])
                    .size()
                )
            self.module.add_submodule(
                sparsify_name, Sparsify(tensor_size, sparseness=sparsify_format)
            )
            prev_node = self.new_graph.create_node(
                "call_module", sparsify_name, args=(prev_node,)
            )

            if self.submodules.get("approximator"):
                prev_node = self.new_graph.create_node(
                    "call_module", "approximator", args=(prev_node,)
                )

        get_attr_node_cast = self.new_graph.create_node(
            "call_module", cast_name, args=(prev_node,)
        )
        return Proxy(get_attr_node_cast, self.tracer)

    def call_function(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``call_function`` node. Insert a cast node after the node.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns
            A Proxy containing the last node added (cast) and the tracer of the new graph
        """
        assert callable(target)

        call_fnc_node = self.new_graph.call_function(target)
        # Observed that inputs to the functions will be wrapped in proxies, parameters of
        # functions is not wrapped in proxies. We need to do a unwrap for proxies before passing to new node.
        call_fnc_node.args = process_args(args)
        call_fnc_node.kwargs = process_kwargs(kwargs)
        # approx_name = call_fnc_node.name + "_approx"
        # self.module.add_submodule(approx_name, Approximator())
        # call_fnc_node_approx = self.new_graph.create_node(
        #     "call_module", approx_name, args=(call_fnc_node,)
        # )
        result_node = call_fnc_node
        layer = self.scopeDict[call_fnc_node.name][0]
        if self.config:
            cast_name = call_fnc_node.name + "_cast"
            layer_key = layer.split("__")[-1]
            if (
                layer_key
                and layer_key in self.config
                and self.config[layer_key]["instance"].lower()
                in call_fnc_node.name.lower()
            ):
                cast_format = self.config[layer_key]["output_format"]
                self.module.add_submodule(cast_name, CastTo(format=cast_format))
                result_node = self.new_graph.create_node(
                    "call_module", cast_name, args=(result_node,)
                )
        return Proxy(result_node, self.tracer)

    def call_method(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """
        Execute a ``call_method`` node.

        Args:
            target (Target): The call target for this node. See
                `Node <https://pytorch.org/docs/master/fx.html#torch.fx.Node>`__ for
                details on semantics
            args (Tuple): Tuple of positional args for this invocation
            kwargs (Dict): Dict of keyword arguments for this invocation

        Returns
            A Proxy containing the call_method node and the tracer of the new graph
        """
        assert isinstance(target, str)
        call_method_node = self.new_graph.call_method(target)
        # Observed that inputs to the functions will be wrapped in proxies, parameters of
        # functions is not wrapped in proxies. We need to do a unwrap for proxies before passing to new node.
        call_method_node.args = process_args(args)
        call_method_node.kwargs = process_kwargs(kwargs)
        return Proxy(call_method_node, self.tracer)



from abc import ABC
from contextlib import contextmanager, ExitStack
from dataclasses import dataclass
from typing import Callable, Optional, List
import torch
from skopt.space import Space
from dmx.compressor.modeling.nn import DmxModule
from dmx.compressor.numerical.observer import (
    HistogramObserver,
    ObserverBase,
)


class DmxBaseRecipe(ABC):
    r"""
    This is an abstract class of ADVANCED mode recipe.
    """

    def __init__(self, hp_gen: Callable, **kwargs):
        super().__init__()
        self.generate_hyperparams = hp_gen
        self.recipe_context_manager = None

    @contextmanager
    def applied_to(self, _model, save_checkpoint_to: Optional[str] = None):
        _hyperparams = self.generate_hyperparams(_model)
        with ExitStack() as stack:
            try:
                yield [
                    stack.enter_context(self.recipe_context_manager(_m, _p))
                    for _m, _p in _hyperparams.items()
                ]
            finally:
                if hasattr(
                    _model, "_save_specific_layers_state_dict_and_register_urls"
                ):
                    _model._save_specific_layers_state_dict_and_register_urls(
                        _hyperparams.keys(), save_checkpoint_to
                    )


@dataclass
class DmxQuantizerCalibrationHyperparams:
    r"""
    Fake quantizer (i.e. CastTo) calibration hyperparameters with default values
    """

    observer_cls: ObserverBase = HistogramObserver
    qscheme_to_overload: torch.qscheme = torch.per_tensor_symmetric
    group_size: Optional[int] = None
    ch_axis: Optional[int] = None


@dataclass
class DmxModuleQuantizerCalibrationHyperparams:
    r"""
    DmxModule boundary cast quantizers calibration hyperparameters with default values
    """

    inputs: Optional[List[DmxQuantizerCalibrationHyperparams]] = None
    outputs: Optional[List[DmxQuantizerCalibrationHyperparams]] = None
    weight: Optional[DmxQuantizerCalibrationHyperparams] = None
    weight_storage: Optional[DmxQuantizerCalibrationHyperparams] = None


@dataclass
class DmxModuleSmoothQuantHyperparams:
    r"""
    DmxModule SmoothQuant hyperparameters with default values
    """

    migration_strength: float = 0.5
    fuse_to_weight: bool = False


@dataclass
class DmxModuleGPTQHyperparams:
    r"""
    DmxModule GPTQ hyperparameters with default values
    """

    microblock_size: int = 1
    block_size: int = 128
    percdamp: float = 0.01


@dataclass
class DmxApproximationFunctionTuningHyperparams:
    r"""
    Approximation function extra_params tuning hyperparameters with default values
    """

    search_space: Optional[Space] = None


@dataclass
class DmxSLaNCHyperparams:
    r"""
    SLaNC hyperparamters with default values
    """

    position: Optional[str] = None  # "post_attn" or "post_mlp" or "first"
    mlp_type: Optional[str] = None # None or "standard" or "llama"
    device: Optional[torch.device] = None # device of layer norm
    prev_ln_weight: Optional[torch.nn.Module] = None # previous layer norm weight
    fc1: Optional[torch.nn.Module] = None # fc1 from previous standard mlp layer
    fc2: Optional[torch.nn.Module] = None # fc2 from previous standard mlp layer
    up_proj: Optional[torch.nn.Module] = None # up_proj from llama mlp layer
    down_proj: Optional[torch.nn.Module] = None # down_proj from llama mlp layer
    gate_proj: Optional[torch.nn.Module] = None # gate_proj from llama mlp layer
    v_proj: Optional[torch.nn.Module] = None # v_proj from (self/cross) attention layer
    o_proj: Optional[torch.nn.Module] = None # o_proj/out_proj from (self/cross) attention layer


class DmxQuantizerCalibrationRecipe(DmxBaseRecipe):
    r"""
    Fake quantizer calibration recipe
    """

    def __init__(self, hp_gen, **kwargs):
        super().__init__(hp_gen, **kwargs)
        self.recipe_context_manager = DmxModule.calibrating_quantizers


class DmxSmoothQuantRecipe(DmxBaseRecipe):
    r"""
    SmoothQuant recipe
    """

    def __init__(self, hp_gen, **kwargs):
        super().__init__(hp_gen, **kwargs)
        self.recipe_context_manager = DmxModule.calibrating_smoothquant


class DmxGPTQRecipe(DmxBaseRecipe):
    r"""
    GPTQ recipe
    """

    def __init__(self, hp_gen, **kwargs):
        super().__init__(hp_gen, **kwargs)
        self.recipe_context_manager = DmxModule.optimal_brain_compressing


class DmxApproximationFunctionTuningRecipe(DmxBaseRecipe):
    r"""
    Approximation function extra_params tuning recipe
    """

    def __init__(self, hp_gen, **kwargs):
        super().__init__(hp_gen, **kwargs)
        self.recipe_context_manager = DmxModule.tuning_approximation_function


class DmxSLaNCRecipe(DmxBaseRecipe):
    r"""
    SLaNC norm tuning for LayerNorm|RMSNorm recipe
    Paper: https://arxiv.org/abs/2410.10553
    """

    def __init__(self, hp_gen, **kwargs):
        super().__init__(hp_gen, **kwargs)
        self.recipe_context_manager = DmxModule.slanc_tuning



from abc import abstractmethod
import time
from typing import Optional,List
from contextlib import contextmanager
from types import SimpleNamespace
import torch
from torch import Tensor
from torch.fx import Graph
import inspect

from dmx.compressor.numerical import NumericalCastMixin, Same, CastTo
from dmx.compressor.plugins import PluginLayerData,PluginBase
from dmx.compressor.sparse import (
    WeightSparseMixin,
    Dense,
    LazySparsify,
)
from dmx.compressor.functional import (
    ApproximationMixin,
    NoApproximation,
)
from dmx.compressor.perf_proxy import PerformanceProxyMixin
from dmx.compressor.layer_reconstruction import LayerReconstructionMixin

import warnings

__ALL__ = ["DmxModuleType", "DmxModule", "DmxModuleConfig", "is_configurable"]


class DmxModuleType(type):
    pass


class DmxModule(
    ApproximationMixin,
    PerformanceProxyMixin,
    LayerReconstructionMixin,
    NumericalCastMixin,
    WeightSparseMixin,
    torch.nn.Module,
    metaclass=DmxModuleType,
):
    r"""
    Extended torch.nn.Module for Dmx to support quantization.

    Args:
        *args (Optional[Tuple]): variable length of args
        state_dict_url (Optional[str]): Url for loading the state dicts. Defaults to None.
        **kwargs (Optional[Dict]): variable length of kwargs

    Attributes:
        state_dict_url (str): Url for loading the module state dicts.
    """

    is_compound = False
    functional_forward = None
    plugins : List[PluginBase] = []
    
    def __init__(self, *args, state_dict_url: Optional[str] = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.state_dict_url = state_dict_url
        if state_dict_url is not None:
            self.load_state_dict_and_register_url(state_dict_url)

    def configure(self, config) -> None:
        """
        A function that changes the format of the ops and loading state dicts according to the config file.

        Args:
            config (DmxModuleConfig): config file for setting new formats and loading state dicts.

        """
        # numerics transformation
        if "input_formats" in config:
            self.input_casts.set_format(format=config["input_formats"])
        if "output_formats" in config:
            self.output_casts.set_format(format=config["output_formats"])
        if self.accum_cast is not None and "accum_format" in config:
            self.accum_cast.set_format(format=config["accum_format"])
        if self.weight_storage_cast is not None and "weight_storage_format" in config:
            self.weight_storage_cast.set_format(format=config["weight_storage_format"])
        if self.weight_cast is not None and "weight_format" in config:
            self.weight_cast.set_format(format=config["weight_format"])
        if self.bias_cast is not None and "bias_format" in config:
            self.bias_cast.set_format(format=config["bias_format"])
        if self.smoothquant is not None and "smoothquant_scale_format" in config:
            self.smoothquant.set_scale_format(format=config["smoothquant_scale_format"])
        # sparsity transformation
        if self.weight_sparsifier is not None and "weight_sparseness" in config:
            self.weight_sparsifier.configure(sparseness=config["weight_sparseness"])
        # custom logic transformation
        if "approximation_function" in config:
            self.approximator.set_function(config["approximation_function"])

        # load state_dict if provided
        if (
            "state_dict_url" in config
            and config["state_dict_url"] != self.state_dict_url
        ):
            self.load_state_dict_and_register_url(config["state_dict_url"])

    transform = configure  # NOTE: to be deprecated

    def load_state_dict_and_register_url(self, url: str) -> None:
        """
        A function that loads state dict from a url and sets url to self.state_dict_url

        Args:
            ulr (str): url for loading the state dict
        """
        self.load_state_dict(torch.hub.load_state_dict_from_url(url))
        self.state_dict_url = url

    def save_state_dict_and_register_url(self, parent_dir: str) -> None:
        """
        A function that saves the current state dict of the module to a url under a specified parent directory

        Args:
            parent_dir (str): parent directory for the url
        """
        from dmx.compressor.utils import save_state_dict_and_return_url

        url = save_state_dict_and_return_url(self, parent_dir)
        self.state_dict_url = url

    def dmx_config(self, freeze=False):
        """
        A function that the DmxModuleConfig object for the module

        Args:
            freeze (bool): if True, both state dict and ops formats would be included in the returned DmxModuleConfig. If False, only state dict will be included.

        Returns:
            A DmxModuleConfig object for the module
        """
        return DmxModuleConfig.from_module(self, freeze)

    def fold_weight_and_bias(self) -> None:
        """
        A function that applies the ops the weights and biases using the corresponding formats.
        """
        with torch.no_grad():
            # bias cast
            if self.bias_cast is not None and not isinstance(self.bias_format, Same):
                self.bias.data = self.bias_cast(self.bias.data)
                self.bias_cast = CastTo(format=Same())
            # weight sparsifier
            if self.weight_sparsifier is not None and not isinstance(
                self.weight_sparseness, Dense
            ):
                self.weight.data = self.effective_weight
                self.weight_sparsifier = LazySparsify(sparseness=Dense())
            # weight smoothquant
            if (
                self.smoothquant is not None
                and self.smoothquant.fused_to_weight[0] == 0
            ):
                self.smoothquant.fuse_to_weight(self.weight)
            # weight storage cast
            if self.weight_storage_cast is not None and not isinstance(
                self.weight_storage_cast, Same
            ):
                self.weight.data = self.weight_storage_cast(self.weight.data)
                self.weight_storage_cast = CastTo(format=Same())
            # weight cast
            if self.weight_cast is not None and not isinstance(self.weight_cast, Same):
                self.weight.data = self.weight_cast(self.weight.data)
                self.weight_cast = CastTo(format=Same())

    @property
    def weight_hypernet(self):
        """
        Returns a function that processes weight according to the ops format of the module
        """

        def _weight_hypernet(_w):
            if self.weight_sparsifier is not None:
                _w = self.weight_sparsifier(_w)
            if (
                self.smoothquant is not None
                and self.smoothquant.fused_to_weight[0] == 0
            ):
                _w = self.smoothquant.scale_weight(_w)
            if self.weight_storage_cast is not None:
                _w = self.weight_storage_cast(_w)
            if self.weight_cast is not None:
                _w = self.weight_cast(_w)
            return _w

        return _weight_hypernet

    @property
    def _weight(self):
        """
        Returns the quantized weights of the module
        """
        return self.weight_hypernet(self.weight)

    @property
    def _bias(self):
        """
        Returns the quantized bias of the module
        """

        return self.bias_cast(self.bias) if self.bias_cast is not None else None

    def forward(self, input: Tensor, *args, **kwargs) -> Tensor:
        """
        Forward pass of the module with quantization ops applied.

        Args:
            input (Tensor): input tensor to be passed through the module
            *args (Optional[Tuple]): variable length of args
            **kwargs (Optional[Dict]): variable length of kwargs
        """
        _dtype, _device = input.dtype, input.device
        if hasattr(self, "weight") and self.weight is not None:
            _device = self.weight.device
        if self.smoothquant is not None:
            if self.smoothquant.dynamic[0] == 1 or self.smoothquant.calibrating:
                self.update_smoothquant_scale(input)
            input = self.smoothquant.scale_input(input)
        _input, args, kwargs = self.input_casts(input, *args, **kwargs)
        if self.obc is not None:
            self.obc.measure_hessian(_input)
        _input, args, kwargs = self.align_device(_input, args, kwargs, _device)
        if self.aft is not None:
            self.aft.optimize(_input, *args, **kwargs)
        _output = self._forward(_input, *args, **kwargs)
        output = self.output_casts(_output, output=True)
        plugin_data = PluginLayerData(input_before_cast = input,
                                      input_after_cast = _input,
                                      output_before_cast = _output,
                                      output_after_cast = output,
                                      mod = self,
                                      args = args,
                                      kwargs = kwargs)
        plugins_copy = DmxModule.plugins.copy()
        for p in plugins_copy:
            #To avoid infinite recursion if the plugin calls forward on the DmxModule
            DmxModule.plugins.remove(p)
            p.process_layer(plugin_data)
            DmxModule.plugins = plugins_copy.copy()
            
        if self.flop_counter_enabled:
            self.count_flops(input, output)
        if self.align_boundary_dtype:
            output = (
                type(output)(a.to(_dtype) for a in output)
                if isinstance(output, (tuple, list))
                else output.to(_dtype)
            )
        return output

    def align_device(self, _input, args, kwargs, _device):
        _input = _input.to(_device) if _input.device != _device else _input
        args = tuple(
            x.to(_device) if isinstance(x, torch.Tensor) and x.device != _device else x
            for x in args
        )
        for k in kwargs.keys():
            if isinstance(kwargs[k], Tensor) and kwargs[k].device != _device:
                kwargs[k] = kwargs[k].to(_device)
        return _input, args, kwargs

    def update_params_with_raw(self, raw: torch.nn.Module) -> None:
        """
        Update parameters of a DmxModule from a torch.nn.Module.

        Args:
            raw (torch.nn.Module): the torch module to copy parameters from.
        """
        self.load_state_dict(raw.state_dict(), assign=True)
        # Inherit device from raw module
        for n, m in raw.named_parameters():
            device = "cuda" if torch.cuda.is_available() else "cpu"
            device = m.device if m.device != "cpu" else device
            setattr(self, n, torch.nn.Parameter(getattr(self, n).to(device)))
        # inherit some module attributes from raw module
        self.training = raw.training
        if hasattr(raw, "dtype"):
            self.dtype = raw.dtype

    @contextmanager
    def monitoring(self, _records: list):
        """
        Context manager for monitoring input/output to/from the DmxModule
        """

        def recorder(_mod, _inp, _input_kwargs, _out) -> None:
            _records.append(
                dict(
                    input=(_inp, _input_kwargs),
                    output=_out,
                )
            )

        _h = self.register_forward_hook(recorder, with_kwargs=True)
        yield self
        _h.remove()

    @contextmanager
    def measuring_runtime(self, _records: list, device: torch.device):
        """
        Context manager for monitoring runtime of DmxModule
        """

        def tick(_mod, _inp):
            assert len(_records) == 0 or not isinstance(
                _records[-1], list
            ), f"multiple tick calls without tock in between: {_records}"
            if device.type == "cuda":
                start_t = torch.cuda.Event(enable_timing=True)
                start_t.record()
            else:
                start_t = time.perf_counter()
            _records.append([start_t])

        def tock(_mod, _inp, _out):
            assert len(_records) > 0 and isinstance(
                _records[-1], list
            ), f"tock called without preceding tick : {_records}"
            # start_t can be float(time) or cuda event
            start_t = _records.pop()[0]
            if device.type == "cuda":
                end_t = torch.cuda.Event(enable_timing=True)
                end_t.record()
                # The synchronization slows things down. We might need to wait till
                # the model finishes running to synchronize and gather the event deltas
                torch.cuda.synchronize()
                _records.append(start_t.elapsed_time(end_t) / 1000)

            else:
                end_t = time.perf_counter()
                _records.append(end_t - start_t)

        _h_pre = self.register_forward_pre_hook(tick)
        _h_post = self.register_forward_hook(tock)
        yield self
        _h_pre.remove()
        _h_post.remove()

    @abstractmethod
    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        raise NotImplementedError("to_compiler_graph not implemented!")


class DmxModuleConfig(dict):
    @classmethod
    def from_module(cls, module: DmxModule, freeze=False):
        """
        A function that stores state and ops format of the module in a DmxModuleConfig object

        Args:
            module (DmxModule): Target module for creating the DmxModuleConfig

        Returns:
            A DmxModuleConfig object that stores state and ops format of the module in a DmxModuleConfig object
        """
        cc = SimpleNamespace()
        cc.instance_of = module.__class__
        if isinstance(module, DmxModule):
            if module.input_formats is not None and (
                freeze
                or not all(isinstance(f, Same) for f in module.input_formats.values())
            ):
                cc.input_formats = module.input_formats
            if module.output_formats is not None and (
                freeze
                or not all(isinstance(f, Same) for f in module.input_formats.values())
            ):
                cc.output_formats = module.output_formats
            if module.accum_format is not None and (
                freeze or not isinstance(module.accum_format, Same)
            ):
                cc.accum_format = module.accum_format
            if module.weight_format is not None and (
                freeze or not isinstance(module.weight_format, Same)
            ):
                cc.weight_format = module.weight_format
            if module.weight_storage_format is not None and (
                freeze or not isinstance(module.weight_storage_format, Same)
            ):
                cc.weight_storage_format = module.weight_storage_format
            if module.bias_format is not None and (
                freeze or not isinstance(module.bias_format, Same)
            ):
                cc.bias_format = module.bias_format
            if module.smoothquant is not None and (
                freeze or not isinstance(module.smoothquant.scale_cast.format, Same)
            ):
                cc.smoothquant_scale_format = module.smoothquant.scale_cast.format
            if module.weight_sparseness is not None and (
                freeze or not isinstance(module.weight_sparseness, Dense)
            ):
                cc.weight_sparseness = module.weight_sparseness
            if freeze or not isinstance(module.approximation_function, NoApproximation):
                cc.approximation_function = module.approximation_function
            if module.state_dict_url is not None:
                cc.state_dict_url = module.state_dict_url
        return cc.__dict__


def is_configurable(m):
    return isinstance(m, DmxModule)


class DmxGraph(Graph):
    def __init__(self, owning_module=None, tracer_cls=None, tracer_extras=None):
        super().__init__(owning_module, tracer_cls, tracer_extras)
        try:
            import dmx.ops
        except:
            warnings.warn("Falling back to dummy q/dq torch ops")

            @torch.library.custom_op("dmx_ops::quantize.Scalar", mutates_args=[])
            def quantize(
                t: float, scale: float, zero_point: float, format: str
            ) -> torch.Tensor:
                return torch.Tensor([t])

            @torch.library.custom_op("dmx_ops::quantize.Tensor", mutates_args=[])
            def quantize(
                t: torch.Tensor, scale: float, zero_point: float, format: str
            ) -> torch.Tensor:
                return t.clone()

            @quantize.register_fake
            def _(
                t: torch.Tensor, scale: float, zero_point: float, format: str
            ) -> torch.Tensor:
                return torch.empty_like(t)

            @torch.library.custom_op("dmx_ops::dequantize.Tensor", mutates_args=[])
            def dequantize(
                t: torch.Tensor, scale: float, zero_point: float
            ) -> torch.Tensor:
                return t.clone()

            @dequantize.register_fake
            def _(t: torch.Tensor, scale: float, zero_point: float) -> torch.Tensor:
                return torch.empty_like(t)

    def qdq_node(self, node, cast_name, cast_format):
        from operator import attrgetter

        if node is None or cast_name is None:
            return node

        _scale = self.get_attr(f"{cast_name}.scale")
        _zero_point = self.get_attr(f"{cast_name}.zero_point")
        _q = self.call_function(
            torch.ops.dmx_ops.quantize,
            (
                node,
                _scale,
                _zero_point,
                cast_format,
            ),
        )
        dq_node = self.call_function(
            torch.ops.dmx_ops.dequantize,
            (_q, _scale, _zero_point),
        )
        return dq_node

    def create_placeholders(self, names, cast_names=None, cast_formats=None):
        placeholder_nodes = []
        if cast_names is None:
            cast_names = [None] * len(names)
            cast_formats = [None] * len(names)
        assert len(cast_names) == len(
            names
        ), "List of placeholder names and cast names shuold have same length!"
        for name, c_name, c_format in zip(names, cast_names, cast_formats):
            _n = self.placeholder(name, c_name, c_format)
            placeholder_nodes.append(_n)
        if len(placeholder_nodes) == 1:
            return placeholder_nodes[0]
        return placeholder_nodes

    def placeholder(
        self,
        name,
        cast_name=None,
        cast_format=None,
        type_expr=None,
        default_value=inspect.Signature.empty,
    ):
        node = super().placeholder(name, type_expr, default_value)
        node = self.qdq_node(node, cast_name, cast_format)
        return node

    def get_attr(
        self,
        qualified_name,
        cast_name=None,
        cast_format=None,
        optional_arg=True,
        type_expr=None,
    ):
        """
        optional_arg: controlling whether None will be returned instead of a Node. e.g. optional_arg = linear_mod.bias will return None when the module does not have a bias term.
                      Defaults to True.
        """
        if optional_arg is None:
            return None
        node = super().get_attr(qualified_name, type_expr)
        node = self.qdq_node(node, cast_name, cast_format)
        return node

    def create_node(
        self,
        op,
        target,
        args=None,
        kwargs=None,
        name=None,
        type_expr=None,
        cast_name=None,
        cast_format=None,
    ):
        node = super().create_node(op, target, args, kwargs, name, type_expr)
        node = self.qdq_node(node, cast_name, cast_format)
        return node

    def call_function(
        self,
        the_function,
        args=None,
        kwargs=None,
        type_expr=None,
        cast_name=None,
        cast_format=None,
    ):
        node = super().call_function(the_function, args, kwargs, type_expr)
        node = self.qdq_node(node, cast_name, cast_format)
        return node

    def call_method(
        self,
        method_name,
        args=None,
        kwargs=None,
        type_expr=None,
        cast_name=None,
        cast_format=None,
    ):
        node = super().call_method(method_name, args, kwargs, type_expr)
        node = self.qdq_node(node, cast_name, cast_format)
        return node



#!/usr/bin/env python3

import torch
from torch import fx, nn
import torch.nn.functional as F
from dmx.compressor.modeling import DmxModel
from dmx.compressor.modeling import nn as dmxnn
from dmx.compressor.fx.transform import substitute_transform
from dmx.compressor.fx.transformer.utils import dmx_aware_mapping

RANDOM_SEED = 0
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.manual_seed(RANDOM_SEED)


def check_all_dmx(gm: torch.nn.Module) -> bool:
    """
    Checking that all modules are dmx modules
    """
    assert isinstance(gm, fx.graph_module.GraphModule), True
    for node in gm.graph.nodes:
        if node.op == "call_module":
            module = gm.get_submodule(node.target)
            node_key = type(module).__module__ + "." + type(module).__name__
            if (
                not isinstance(module, dmxnn.DmxModule)
                and node_key in dmx_aware_mapping
            ):
                return False
    return True


class Lenet5(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 6, 5)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)
        self.mp1 = nn.MaxPool2d((2, 2))
        self.mp2 = nn.MaxPool2d(2)

    def forward(self, x):
        x = self.mp1(F.relu(self.conv1(x)))
        x = self.mp2(F.relu(self.conv2(x)))
        x = torch.flatten(x, 1)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


def test_lenet5():
    net = Lenet5()
    gm_before = fx.symbolic_trace(net)
    gm = DmxModel.from_torch(net)
    inp = torch.rand(1, 95, 95)
    output_before = gm_before(inp)
    output = gm(inp)
    assert check_all_dmx(gm._gm), True
    assert (output_before - output).abs().sum() <= 1e-5, True


class MultiInputNet(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.linear_x = nn.Linear(32, 64)
        self.linear_y = nn.Linear(64, 64)
        self.linear_z = nn.Linear(16, 64)

    def forward(self, x, y, z):
        return self.linear_x(x) + self.linear_y(y) + self.linear_z(z)


def test_multiple_inputs():
    net = MultiInputNet()
    gm_before = fx.symbolic_trace(net)
    gm = DmxModel.from_torch(net)
    x = torch.rand(8, 32)
    y = torch.rand(8, 64)
    z = torch.rand(8, 16)
    output_before = gm_before(x, y, z)
    output = gm(x, y, z)
    assert check_all_dmx(gm._gm), True
    assert (output_before - output).abs().sum() <= 1e-5, True


class Add(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.iden = nn.Identity()

    def forward(self, x, y):
        return self.iden(x) + y


class ResConnection(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.linear_1 = nn.Linear(32, 64)
        self.relu = nn.ReLU()
        self.linear_2 = nn.Linear(64, 32)
        self.add = Add()

    def forward(self, x):
        return self.add(self.linear_2(self.relu(self.linear_1(x))), x)


def test_res_connection():
    net = ResConnection()
    gm_before = fx.symbolic_trace(net)
    gm = DmxModel.from_torch(net)
    x = torch.rand(8, 32)
    output_before = gm_before(x)
    output = gm(x)
    assert check_all_dmx(gm._gm), True
    assert (output_before - output).abs().sum() <= 1e-5, True



import pytest
import torch
import transformers
import dmx
from dmx.compressor.modeling import nn as dmxnn

RANDOM_SEED = 0

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
torch.manual_seed(RANDOM_SEED)


@pytest.mark.parametrize("bsz", [1, 8])
@pytest.mark.parametrize("dim", [64, 128])
@pytest.mark.parametrize("seq_len", [64,512])
@pytest.mark.parametrize("nheads", [4, 8,32])
@pytest.mark.parametrize("unsqueeze_dim", [1,2])
def test_apply_rope(bsz, dim,seq_len,nheads,unsqueeze_dim):
    assert unsqueeze_dim in [1,2]
    if unsqueeze_dim == 1:
        q = torch.rand(bsz,nheads,seq_len,dim).to(device)
        k = torch.rand(bsz,nheads,seq_len,dim).to(device)        
    elif unsqueeze_dim == 2:
        q = torch.rand(bsz,seq_len,nheads,dim).to(device)
        k = torch.rand(bsz,seq_len,nheads,dim).to(device)        

    cos = torch.rand(bsz,seq_len,dim // 2).repeat(1,1,2).to(device)
    sin = torch.rand(bsz,seq_len,dim // 2).repeat(1,1,2).to(device)

    
    gt_q_pos,gt_k_pos = transformers.models.llama.modeling_llama.apply_rotary_pos_emb(
        q,k,cos,sin,unsqueeze_dim = unsqueeze_dim)

    dmx_rope = dmxnn.ApplyRotaryPosEmb()
    dmx_q_pos,dmx_k_pos = dmx_rope(q,k,cos,sin,unsqueeze_dim = unsqueeze_dim)

    rope_config = [x for x in dmx.compressor.config_rules.BASIC if \
                   type(dmx_rope) in x.module_types][0].module_config
    dmx_rope.configure(rope_config)

    basic_q_pos,basic_k_pos = dmx_rope(q,k,cos,sin,unsqueeze_dim = unsqueeze_dim)
    

    atol = 1e-12


    assert torch.allclose(gt_q_pos, dmx_q_pos, atol=atol)
    assert torch.allclose(gt_k_pos, dmx_k_pos, atol=atol)    

    atol = 1e-2    
    assert torch.allclose(gt_q_pos, basic_q_pos, atol=atol)
    assert torch.allclose(gt_k_pos, basic_k_pos, atol=atol)    
    




#!/usr/bin/env python3

from collections import OrderedDict

import torch
from torch.library import custom_op

from dmx.compressor.modeling import DmxModel, DmxConfigRule
from dmx.compressor.modeling import nn as dmxnn
from dmx.compressor.numerical import CastTo, CastToDict


@custom_op("mylib::my_func", mutates_args=())
def my_func(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    out = x + y
    return out


class DmxSubmod(dmxnn.DmxModule):
    def __init__(self) -> None:
        super().__init__()
        self.input_casts = CastToDict(
            OrderedDict({"x_cast": CastTo(), "y_cast": CastTo()})
        )

    def _forward(self, x, y):
        out = x + y
        return out


class Model(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        out = my_func(x, x)
        return out


def test_additional_mappings():
    model = Model()
    x = torch.ones((1, 1))
    additional_mappings = {"torch.ops.mylib.my_func.default": DmxSubmod}
    dmx_model = DmxModel.from_torch(model, additional_mappings)
    dmx_model(x)

    assert isinstance(dmx_model._gm.my_func_default, DmxSubmod)


def test_no_additional_mappings():
    model = Model()
    x = torch.ones((1, 1))
    dmx_model = DmxModel.from_torch(model)
    dmx_model(x)

    assert hasattr(dmx_model._gm, "my_func_default") == False



from collections import OrderedDict
import torch
from torch import Tensor, Size
from torch.fx import Graph, symbolic_trace
import transformers
import transformers.activations

from dmx.compressor.numerical import Same, CastTo, CastToDict
from . import DmxModule
from .torch_modules import GELUBase
from .core import DmxGraph


class GemmaRMSNorm(DmxModule, transformers.models.gemma.modeling_gemma.GemmaRMSNorm):
    r"""
    An extension of RMSNorm layer to support DmxModule configurations.
    This module performs RMS-based layer normalization on the input tensor.
    The layer normalization is characterized by the `hidden_size` and an optional `eps` value for numerical stability.

    Args:
        dim (int): The size of the hidden layer (number of hidden units).
        eps (float, optional): A small constant added to the denominator for numerical stability. Defaults to 1e-6.

    Methods:
        _forward (_input: Tensor) -> Tensor: Computes the forward pass of the RMS layer normalization.
    """

    def __init__(
        self,
        dim: int,
        eps: float = 1e-6,
    ) -> None:
        super().__init__(dim, eps=eps)

    def _norm(self, x, eps):
        return x * torch.rsqrt(x.pow(2).mean(-1, keepdim=True) + eps)

    def functional_forward(self, x, weight, eps):
        output = self._norm(x.float(), eps)
        output = output * (1.0 + weight.float())
        return output.type_as(x)

    def _forward(self, _input: Tensor) -> Tensor:
        _output = self.approx_forward((_input,), self._weight, self.eps)
        return _output

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        r"""
        Creates a new RMSNorm object (DmxModule) from a given PyTorch RMSNorm layer.

        Args:
            raw (torch.nn.Module): A PyTorch RMSNorm layer to be converted.

        Returns:
            DmxModule: A RMSNorm object that has the same configuration as the input PyTorch RMSNorm layer.
        """
        initial_dmx = cls(
            dim=raw.weight.shape[0],
            eps=raw.variance_epsilon if hasattr(raw, "variance_epsilon") else raw.eps,
        )
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            _weight_dq = g.get_attr(
                "weight", "weight_cast", repr(self.weight_cast.format)
            )

            # Non Tensor Attributes (no need to quantize)
            eps = g.get_attr("eps")

            args = ((_input_dq), _weight_dq, eps)
            _output_dq = g.create_node(
                "call_function",
                self.functional_forward,
                args,
                name="GemmanRMSNorm",
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class NewGELU(GELUBase):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(transformers.activations.NewGELUActivation, *args, **kwargs)

    def functional_forward(self, x):
        return transformers.activations.NewGELUActivation().forward(x)


class FastGELU(GELUBase):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(transformers.activations.FastGELUActivation, *args, **kwargs)

    def functional_forward(self, x):
        return transformers.activations.FastGELUActivation().forward(x)


class QuickGELU(GELUBase):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(transformers.activations.QuickGELUActivation, *args, **kwargs)

    def functional_forward(self, x):
        return transformers.activations.QuickGELUActivation().forward(x)


class ClippedGELU(GELUBase):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            transformers.activations.ClippedGELUActivation, *args, **kwargs
        )

    def functional_forward(self, x):
        return transformers.activations.ClippedGELUActivation(
            self.min, self.max
        ).forward(x)


class BloomGELU(GELUBase):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            transformers.models.bloom.modeling_bloom.BloomGelu, *args, **kwargs
        )

    def functional_forward(self, x):
        return transformers.models.bloom.modeling_bloom.BloomGelu().forward(x)


class ApplyRotaryPosEmbBase(torch.nn.Module):
    def rotate_half(self, x):
        """Rotates half the hidden dims of the input."""
        x1 = x[..., : x.shape[-1] // 2]
        x2 = x[..., x.shape[-1] // 2 :]
        return torch.cat((-x2, x1), dim=-1)

    def forward(self, q, k, cos, sin, unsqueeze_dim=1):
        """Applies Rotary Position Embedding to the query and key tensors.

        Args:
            q (`torch.Tensor`): The query tensor.
            k (`torch.Tensor`): The key tensor.
            cos (`torch.Tensor`): The cosine part of the rotary embedding.
            sin (`torch.Tensor`): The sine part of the rotary embedding.
            unsqueeze_dim (`int`, *optional*, defaults to 1):
                The 'unsqueeze_dim' argument specifies the dimension along which to unsqueeze cos[position_ids] and
                sin[position_ids] so that they can be properly broadcasted to the dimensions of q and k. For example, note
                that cos[position_ids] and sin[position_ids] have the shape [batch_size, seq_len, head_dim]. Then, if q and
                k have the shape [batch_size, heads, seq_len, head_dim], then setting unsqueeze_dim=1 makes
                cos[position_ids] and sin[position_ids] broadcastable to the shapes of q and k. Similarly, if q and k have
                the shape [batch_size, seq_len, heads, head_dim], then set unsqueeze_dim=2.
        Returns:
            `tuple(torch.Tensor)` comprising of the query and key tensors rotated using the Rotary Position Embedding.
        """

        cos = cos.unsqueeze(unsqueeze_dim)
        sin = sin.unsqueeze(unsqueeze_dim)
        q_embed = (q * cos) + (self.rotate_half(q) * sin)
        k_embed = (k * cos) + (self.rotate_half(k) * sin)
        return q_embed, k_embed


class ApplyRotaryPosEmb(DmxModule, ApplyRotaryPosEmbBase):
    def __init__(self) -> None:
        super().__init__()
        self.input_casts = CastToDict(
            OrderedDict(
                {
                    "q_cast": CastTo(),
                    "k_cast": CastTo(),
                    "cos_cast": CastTo(),
                    "sin_cast": CastTo(),
                }
            )
        )
        self.output_casts = CastToDict(
            OrderedDict({"q_embed_cast": CastTo(), "k_embed_cast": CastTo()})
        )

    def _forward(self, q, k, cos, sin, unsqueeze_dim=1) -> Tensor:
        _output = self.approx_forward((q, k, cos, sin, unsqueeze_dim))
        return _output

    def to_compiler_graph(self):
        # combine these two lines into one context manager
        g = DmxGraph()
        with g.inserting_after():
            # a function that insert placeholder nodes
            dq_q, dq_k, dq_cos, dq_sin = g.create_placeholders(
                ["q", "k", "cos", "sin"],
                [
                    "input_casts.q_cast",
                    "input_casts.k_cast",
                    "input_casts.cos_cast",
                    "input_casts.sin_cast",
                ],
                [
                    repr(self.input_casts.q_cast.format),
                    repr(self.input_casts.k_cast.format),
                    repr(self.input_casts.cos_cast.format),
                    repr(self.input_casts.sin_cast.format),
                ],
            )
            q_res = g.call_function(
                apply_rotary_embeddings,
                (dq_q, dq_cos, dq_sin),
                cast_name="output_casts.q_embed_cast",
                cast_format=repr(self.output_casts.q_embed_cast.format),
            )
            k_res = g.call_function(
                apply_rotary_embeddings,
                (dq_k, dq_cos, dq_sin),
                cast_name="output_casts.k_embed_cast",
                cast_format=repr(self.output_casts.k_embed_cast.format),
            )

            g.output((q_res, k_res))
        return g


def apply_rotary_embeddings(x, cos_embedding, sin_embedding):
    cos = cos_embedding.unsqueeze(1)
    sin = sin_embedding.unsqueeze(1)
    x_embed = (x * cos) + (ApplyRotaryPosEmbBase().rotate_half(x) * sin)
    return x_embed


class LlamaRotaryEmbedding(
    DmxModule, transformers.models.llama.modeling_llama.LlamaRotaryEmbedding
):
    def __init__(self, config, device=None):
        super().__init__(config, device)
        self.cos = None
        self.sin = None
        self.input_casts = CastToDict(
            OrderedDict(
                {
                    "x_cast": CastTo(),
                    "position_ids_cast": CastTo(),
                }
            )
        )
        self.output_casts = CastToDict(
            OrderedDict({"cos_cast": CastTo(), "sin_cast": CastTo()})
        )

    def _forward(self, x: Tensor, cos: Tensor, sin: Tensor) -> Tensor:
        _output = self.approx_forward((x, cos, sin))
        return _output

    def _forward(self, x, position_ids):
        if "dynamic" in self.rope_type:
            self._dynamic_frequency_update(position_ids, device=x.device)

        # Core RoPE block
        inv_freq_expanded = (
            self.inv_freq[None, :, None].float().expand(position_ids.shape[0], -1, 1)
        )
        position_ids_expanded = position_ids[:, None, :].float()
        # Force float32 (see https://github.com/huggingface/transformers/pull/29285)
        device_type = x.device.type
        device_type = (
            device_type
            if isinstance(device_type, str) and device_type != "mps"
            else "cpu"
        )
        with torch.autocast(device_type=device_type, enabled=False):
            freqs = (
                inv_freq_expanded.float() @ position_ids_expanded.float()
            ).transpose(1, 2)
            emb = torch.cat((freqs, freqs), dim=-1)
            cos = emb.cos()
            sin = emb.sin()

        # Advanced RoPE types (e.g. yarn) apply a post-processing scaling factor, equivalent to scaling attention
        cos = cos * self.attention_scaling
        sin = sin * self.attention_scaling

        self.sin = sin.to(dtype=x.dtype)
        self.cos = cos.to(dtype=x.dtype)
        return self.cos, self.sin

    @classmethod
    def from_raw(cls, raw: torch.nn.Module) -> DmxModule:
        initial_dmx = cls(raw.config, raw.inv_freq.device)
        initial_dmx.update_params_with_raw(raw)
        return initial_dmx

    def to_compiler_graph(self) -> Graph:
        if self.sin is None:
            raise ValueError(
                "forward of LlamaROPE must be called before to_compiler_graph"
            )

        g = DmxGraph()
        with g.inserting_after():
            g.create_placeholders(["x", "position_ids"])
            sin = g.get_attr("sin")
            cos = g.get_attr("cos")
            g.output((cos, sin))
        return g



import torch
from torch.fx import Graph
from . import DmxModule, Conv1d as _Conv1d, Conv2d as _Conv2d
import operator
from .core import DmxGraph


class Conv1dUnfold(_Conv1d):
    r"""
    This is an alternative version of the DmxModule .nn.Conv1d,
    without calling torch.nn.functional.conv1d(), but torch.nn.functional.unfold() and torch.matmul() instead.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        **kwargs,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            **kwargs,
        )
        self.input_casts.input_cast.block_dim = 1
        self.weight_cast.block_dim = 1
        if self.bias_cast is not None:
            self.bias_cast.block_dim = -1

    def _forward(self, _input: torch.Tensor) -> torch.Tensor:
        _weight = self._weight.reshape((self.out_channels, -1))
        _input = torch.nn.functional.unfold(
            _input.unsqueeze(-1),
            kernel_size=self.kernel_size + (1,),
            dilation=self.dilation + (1,),
            padding=self.padding + (0,),
            stride=self.stride + (1,),
        )
        _convolution = self.accum_cast(torch.matmul(_weight, _input))
        if self.bias is not None:
            _output = torch.add(_convolution, self._bias.unsqueeze(-1))
        else:
            _output = _convolution
        return _output

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            # PLACEHOLDERS
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            _unsqueeze = g.call_function(torch.unsqueeze, (_input_dq, -1))
            _unfold = g.create_node(
                "call_function",
                torch.nn.functional.unfold,
                (_unsqueeze, self.kernel_size + (1,)),
                dict(
                    dilation=self.dilation + (1,),
                    padding=self.padding + (0,),
                    stride=self.stride + (1,),
                ),
                name="_unfold",
            )

            # _weight
            _weight_storage_dq = g.get_attr(
                "weight", "weight_storage_cast", repr(self.weight_storage_cast.format)
            )
            _weight_dq = g.qdq_node(
                _weight_storage_dq, "weight_cast", repr(self.weight_cast.format)
            )

            _reshape = g.call_function(
                torch.reshape, (_weight_dq, (self.out_channels, -1))
            )
            _matmul_dq = g.call_function(
                torch.matmul,
                (_reshape, _unfold),
                cast_name="accum_cast",
                cast_format=repr(self.accum_cast.format),
            )

            if self.bias is not None:
                _bias_dq = g.get_attr(
                    "bias",
                    "bias_cast",
                    repr(getattr(self.bias_cast, "format", None)),
                )
                _bias_unsqueeze = g.call_function(torch.unsqueeze, (_bias_dq, -1))

                _output = g.call_function(
                    torch.add,
                    (_matmul_dq, _bias_unsqueeze),
                )
            else:
                _output = _matmul_dq

            _output_dq = g.qdq_node(
                _output,
                "output_casts.output_cast",
                repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Conv1dScatter(_Conv1d):
    r"""
    This is an alternative version of the DmxModule .nn.Conv1d,
    without calling torch.nn.functional.conv1d(), but torch.scatter() and torch.matmul() instead.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        **kwargs,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            **kwargs,
        )
        self.input_casts.input_cast.block_dim = 1
        self.weight_cast.block_dim = 1
        if self.bias_cast is not None:
            self.bias_cast.block_dim = -1

    def _forward(self, _input: torch.Tensor) -> torch.Tensor:
        _N, _, _l_in = _input.shape
        _padded_l_in = _l_in + 2 * self.padding[0]
        _l_out = (_padded_l_in - self.kernel_size[0]) // self.stride[0] + 1

        # zero padded input
        _matmul_input = _input.transpose(1, 2).unsqueeze(2)
        _pad = _input.new_zeros(_N, self.padding[0], 1, self.in_channels)
        _matmul_input = torch.cat((_pad, _matmul_input, _pad), 1)

        _matmul_weight = _input.new_zeros(
            self.out_channels, _padded_l_in, self.in_channels, _l_out
        )

        # Create a single weight matrix from all kernels
        _weight_ref = self._weight.transpose(1, 2).unsqueeze(3).repeat(1, 1, 1, _l_out)

        # Create scatter indices
        _indices = _input.new_ones(_l_out, dtype=torch.int64).cumsum(0) - 1
        _indices = _indices * self.stride[0]
        _offset = _input.new_ones(self.kernel_size[0], dtype=torch.int64).cumsum(0) - 1
        _indices = _indices[:, None].repeat(1, self.kernel_size[0]) + _offset
        _indices = _indices.t()
        _indices = _indices.unsqueeze(0).unsqueeze(2).expand_as(_weight_ref)

        # Copy the weight matrix to correct indices
        _matmul_weight.scatter_(dim=1, index=_indices, src=_weight_ref)

        _output = _input.new_zeros(_N, 0, _l_out)
        # OUT_CH is the output_channel size, and is constant
        for Cout in range(self.out_channels):
            # matmul broadcast batch dim corresponds to input batch size, so Cout dim is covered manually
            # torch.sum is reduction over padded sequence
            _sum = torch.sum(
                torch.matmul(_matmul_input, _matmul_weight[Cout])[:, :, 0, :], dim=1
            )
            if self.bias is not None:
                _sum = torch.add(_sum, self._bias[Cout])
            _output = torch.cat((_output, _sum.unsqueeze(1)), 1)
        return _output

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        g = DmxGraph()
        with g.inserting_after():
            # PLACEHOLDERS
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )

            getattr_1 = g.call_function(getattr, (_input_dq, "shape"), {})
            getitem = g.call_function(operator.getitem, (getattr_1, 0), {})
            getitem_1 = g.call_function(operator.getitem, (getattr_1, 1), {})
            getitem_2 = g.call_function(operator.getitem, (getattr_1, 2), {})
            add = g.call_function(operator.add, (getitem_2, 2 * self.padding[0]), {})
            sub = g.call_function(operator.sub, (add, self.kernel_size[0]), {})
            floordiv = g.call_function(operator.floordiv, (sub, self.stride[0]), {})
            add_1 = g.call_function(operator.add, (floordiv, 1), {})
            transpose = g.call_method("transpose", (_input_dq, 1, 2), {})
            unsqueeze = g.call_method("unsqueeze", (transpose, 2), {})
            new_zeros = g.call_method(
                "new_zeros",
                (_input_dq, getitem, self.padding[0], 1, self.in_channels),
                {},
            )
            cat = g.call_function(torch.cat, ((new_zeros, unsqueeze, new_zeros), 1), {})
            new_zeros_1 = g.call_method(
                "new_zeros",
                (_input_dq, self.out_channels, add, self.in_channels, add_1),
                {},
            )

            # _weight
            _weight_storage_dq = g.get_attr(
                "weight", "weight_storage_cast", repr(self.weight_storage_cast.format)
            )
            _weight_dq = g.qdq_node(
                _weight_storage_dq, "weight_cast", repr(self.weight_cast.format)
            )
            transpose_1 = g.call_method("transpose", (_weight_dq, 1, 2), {})
            unsqueeze_1 = g.call_method("unsqueeze", (transpose_1, 3), {})
            repeat = g.call_method("repeat", (unsqueeze_1, 1, 1, 1, add_1), {})
            new_ones = g.call_method(
                "new_ones", (_input_dq, add_1), {"dtype": torch.int64}
            )
            cumsum = g.call_method("cumsum", (new_ones, 0), {})
            sub_1 = g.call_function(operator.sub, (cumsum, 1), {})
            mul = g.call_function(operator.mul, (sub_1, self.stride[0]), {})
            new_ones_1 = g.call_method(
                "new_ones", (_input_dq, self.kernel_size[0]), {"dtype": torch.int64}
            )
            cumsum_1 = g.call_method("cumsum", (new_ones_1, 0), {})
            sub_2 = g.call_function(operator.sub, (cumsum_1, 1), {})
            getitem_3 = g.call_function(
                operator.getitem, (mul, (slice(None, None, None), None)), {}
            )
            repeat_1 = g.call_method("repeat", (getitem_3, 1, self.kernel_size[0]), {})
            add_2 = g.call_function(operator.add, (repeat_1, sub_2), {})
            t = g.call_method("t", (add_2,), {})
            unsqueeze_2 = g.call_method("unsqueeze", (t, 0), {})
            unsqueeze_3 = g.call_method("unsqueeze", (unsqueeze_2, 2), {})
            expand_as = g.call_method("expand_as", (unsqueeze_3, repeat), {})
            scatter_ = g.call_method(
                "scatter_",
                (new_zeros_1,),
                {"dim": 1, "index": expand_as, "src": repeat},
            )
            new_zeros_2 = g.call_method("new_zeros", (_input_dq, getitem, 0, add_1), {})

            for Cout in range(self.out_channels):
                getitem_4 = g.call_function(operator.getitem, (scatter_, Cout), {})
                matmul = g.call_function(operator.matmul, (cat, getitem_4), {})
                getitem_5 = g.call_function(
                    operator.getitem,
                    (
                        matmul,
                        (
                            slice(None, None, None),
                            slice(None, None, None),
                            0,
                            slice(None, None, None),
                        ),
                    ),
                    {},
                )
                sum_1 = g.call_function(torch.sum, (getitem_5,), {"dim": 1})
                if self.bias is not None:
                    _bias_dq = g.get_attr(
                        "bias",
                        "bias_cast",
                        repr(getattr(self.bias_cast, "format", None)),
                    )
                    getitem_6 = g.call_function(operator.getitem, (_bias_dq, Cout), {})
                    add_3 = g.call_function(torch.add, (sum_1, getitem_6), {})
                    sum_1 = add_3
                unsqueeze_4 = g.call_method("unsqueeze", (sum_1, 1), {})
                cat_1 = g.call_function(torch.cat, ((new_zeros_2, unsqueeze_4), 1), {})
                new_zeros_2 = cat_1

            _output_dq = g.qdq_node(
                cat_1,
                "output_casts.output_cast",
                repr(self.output_casts.output_cast.format),
            )

            g.output(_output_dq)
        return g


class Conv2dUnfold(_Conv2d):
    r"""
    This is an alternative version of the DmxModule.nn.Conv2d,
    without calling torch.nn.functional.conv2d(), but torch.nn.functional.unfold() and torch.matmul() instead.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        **kwargs,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            **kwargs,
        )
        self.input_casts.input_cast.block_dim = 1
        self.weight_cast.block_dim = 1
        if self.bias_cast is not None:
            self.bias_cast.block_dim = -1

    def _forward(self, _input: torch.Tensor) -> torch.Tensor:
        _, _, in_height, in_width = _input.shape
        _h_out = (
            in_height + 2 * self.padding[0] - (self.kernel_size[0] - 1) - 1
        ) // self.stride[0] + 1
        _w_out = (
            in_width + 2 * self.padding[1] - (self.kernel_size[1] - 1) - 1
        ) // self.stride[1] + 1
        _weight = self._weight.reshape((self.out_channels, -1))
        _input = torch.nn.functional.unfold(
            _input,
            kernel_size=self.kernel_size,
            dilation=self.dilation,
            padding=self.padding,
            stride=self.stride,
        )
        _convolution = self.accum_cast(torch.matmul(_weight, _input))
        if self.bias is not None:
            _output = torch.add(_convolution, self._bias.unsqueeze(-1))
        else:
            _output = _convolution
        return torch.nn.functional.fold(_output, (_h_out, _w_out), (1, 1))

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        import operator

        g = DmxGraph()
        with g.inserting_after():
            # PLACEHOLDERS
            _input_dq = g.create_placeholders(
                ["_input"],
                ["input_casts.input_cast"],
                [repr(self.input_casts.input_cast.format)],
            )
            getattr_1 = g.call_function(getattr, (_input_dq, "shape"), {})
            getitem_2 = g.call_function(operator.getitem, (getattr_1, 2), {})
            getitem_3 = g.call_function(operator.getitem, (getattr_1, 3), {})
            add = g.call_function(operator.add, (getitem_2, 2 * self.padding[0]), {})
            sub = g.call_function(operator.sub, (add, self.kernel_size[0] - 1), {})
            sub_1 = g.call_function(operator.sub, (sub, 1), {})
            floordiv = g.call_function(operator.floordiv, (sub_1, self.stride[0]), {})
            add_1 = g.call_function(operator.add, (floordiv, 1), {})
            add_2 = g.call_function(operator.add, (getitem_3, 2 * self.padding[1]), {})
            sub_2 = g.call_function(operator.sub, (add_2, self.kernel_size[1] - 1), {})
            sub_3 = g.call_function(operator.sub, (sub_2, 1), {})
            floordiv_1 = g.call_function(operator.floordiv, (sub_3, self.stride[1]), {})
            add_3 = g.call_function(operator.add, (floordiv_1, 1), {})

            # _weight
            _weight_storage_dq = g.get_attr(
                "weight", "weight_storage_cast", repr(self.weight_storage_cast.format)
            )
            _weight_dq = g.qdq_node(
                _weight_storage_dq, "weight_cast", repr(self.weight_cast.format)
            )

            getattr_2 = g.call_function(getattr, (_weight_dq, "device"), {})
            to = g.call_method("to", (_weight_dq, getattr_2), {})
            clone = g.call_method("clone", (to,), {})
            getattr_3 = g.call_function(getattr, (to, "dtype"), {})
            to_1 = g.call_method("to", (clone, getattr_3), {})
            reshape = g.call_method("reshape", (to_1, (self.out_channels, -1)), {})
            unfold = g.call_function(
                torch.nn.functional.unfold,
                (_input_dq, self.kernel_size),
                {
                    "dilation": self.dilation,
                    "padding": self.padding,
                    "stride": self.stride,
                },
            )
            _matmul_dq = g.call_function(
                torch.matmul,
                (reshape, unfold),
                {},
                cast_name="accum_cast",
                cast_format=repr(self.accum_cast.format),
            )
            clone_1 = g.call_method("clone", (_matmul_dq,), {})
            getattr_4 = g.call_function(getattr, (_matmul_dq, "dtype"), {})
            to_2 = g.call_method("to", (clone_1, getattr_4), {})
            if self.bias is not None:
                _bias_dq = g.get_attr(
                    "bias",
                    "bias_cast",
                    repr(getattr(self.bias_cast, "format", None)),
                )
                clone_2 = g.call_method("clone", (_bias_dq,), {})
                getattr_5 = g.call_function(getattr, (_bias_dq, "dtype"), {})
                to_3 = g.call_method("to", (clone_2, getattr_5), {})
                unsqueeze = g.call_method("unsqueeze", (to_3, -1), {})
                add_4 = g.call_function(torch.add, (to_2, unsqueeze), {})
                output = add_4
            else:
                output = to_2
            _output_dq = g.call_function(
                torch.nn.functional.fold,
                (output, (add_1, add_3), (1, 1)),
                cast_name="output_casts.output_cast",
                cast_format=repr(self.output_casts.output_cast.format),
            )
            g.output(_output_dq)
        return g


class Conv2dGather(_Conv2d):
    r"""
    This is an alternative version of the DmxModule.nn.Conv2d,
    without calling torch.nn.functional.conv2d(), but torch.gather() and torch.matmul() instead.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode="zeros",
        **kwargs,
    ) -> None:
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode=padding_mode,
            **kwargs,
        )
        self.input_casts.input_cast.block_dim = 1
        self.weight_cast.block_dim = 1
        if self.bias_cast is not None:
            self.bias_cast.block_dim = -1

    def _forward(self, _input: torch.Tensor) -> torch.Tensor:
        _N, _, in_height, in_width = _input.shape
        _h_out = (
            in_height + 2 * self.padding[0] - (self.kernel_size[0] - 1) - 1
        ) // self.stride[0] + 1
        _w_out = (
            in_width + 2 * self.padding[1] - (self.kernel_size[1] - 1) - 1
        ) // self.stride[1] + 1

        # zero padded input
        _pad_h = _input.new_zeros(_N, self.in_channels, self.padding[0], in_width)
        _padded_input = torch.cat((_pad_h, _input, _pad_h), 2)
        _pad_w = _input.new_zeros(
            _N, self.in_channels, in_height + 2 * self.padding[0], self.padding[1]
        )
        _padded_input = torch.cat((_pad_w, _padded_input, _pad_w), 3)
        _input_ref = (
            _padded_input.unsqueeze(1).unsqueeze(2).repeat(1, _h_out, _w_out, 1, 1, 1)
        )
        del _pad_h, _pad_w, _padded_input

        # Create gather indices for Height dim
        _indices_h = _input.new_ones(_h_out, dtype=torch.int64).cumsum(0) - 1
        _indices_h = _indices_h * self.stride[0]
        _offset_h = (
            _input.new_ones(self.kernel_size[0], dtype=torch.int64).cumsum(0) - 1
        )
        _indices_h = _indices_h[:, None].repeat(1, self.kernel_size[0]) + _offset_h
        _indices_h = (
            _indices_h.unsqueeze(0).unsqueeze(2).repeat(_N, 1, self.in_channels, 1)
        )
        _indices_h = (
            _indices_h.unsqueeze(2)
            .unsqueeze(5)
            .repeat(1, 1, _w_out, 1, 1, in_width + 2 * self.padding[1])
        )

        # Gather Matmul inputs in Height dim
        _matmul_input = _input_ref.gather(dim=4, index=_indices_h)
        del _indices_h, _offset_h, _input_ref

        # Create gather indices for Width dim
        _indices_w = _input.new_ones(_w_out, dtype=torch.int64).cumsum(0) - 1
        _indices_w = _indices_w * self.stride[1]
        _offset_w = (
            _input.new_ones(self.kernel_size[1], dtype=torch.int64).cumsum(0) - 1
        )
        _indices_w = _indices_w[:, None].repeat(1, self.kernel_size[1]) + _offset_w
        _indices_w = (
            _indices_w.unsqueeze(0).unsqueeze(2).repeat(_N, 1, self.in_channels, 1)
        )
        _indices_w = (
            _indices_w.unsqueeze(1)
            .unsqueeze(4)
            .repeat(1, _h_out, 1, 1, self.kernel_size[0], 1)
        )

        # Gather Matmul inputs in Width dim
        _matmul_input = _matmul_input.gather(dim=5, index=_indices_w)
        del _indices_w, _offset_w
        _matmul_input = _matmul_input.view(
            -1, _matmul_input.size(3) * _matmul_input.size(4) * _matmul_input.size(5)
        )

        # Create matmul weight from original weights
        _matmul_weight = self._weight.reshape(
            -1, self._weight.size(1) * self._weight.size(2) * self._weight.size(3)
        ).transpose(0, 1)

        # Matmul calculation
        _output = (
            torch.matmul(_matmul_input, _matmul_weight)
            .view(_N, _h_out, _w_out, -1)
            .transpose(2, 3)
            .transpose(1, 2)
        )
        if self.bias is not None:
            _output = torch.add(
                _output, self._bias.unsqueeze(0).unsqueeze(2).unsqueeze(3)
            )

        return _output

    def to_compiler_graph(self) -> Graph:
        """
        Returns a compiler friendly graph
        """
        raise NotImplementedError("to_compiler_graph not implemented!")



from typing import Union, Optional, Callable
from parse import parse
from bidict import bidict
import torch
import torch.nn as nn
from .functions import *
import transformers

try:
    from dmx.common.vsimd.ported.operator import functional as vsimd

    VSIMD_OP_REF_AVAILABLE = True
except ModuleNotFoundError:
    VSIMD_OP_REF_AVAILABLE = False

torch_function_mapping = bidict(
    {
        "GELU": torch.nn.functional.gelu,
        "SILU": torch.nn.functional.silu,
        "RMS_NORM": torch.nn.functional.rms_norm,
        "LAYER_NORM": torch.nn.functional.layer_norm,
        "SOFTMAX": torch.nn.functional.softmax,
        "EXP": torch.exp,
    }
)

custom_function_mapping = bidict(
    {
        "QUICK_GELU": (
            "quick_gelu",
            transformers.activations.QuickGELUActivation.forward,
        ),
        "APPLY_LLAMA_ROPE": (
            "apply_rotary_pos_emb",
            transformers.models.llama.modeling_llama.apply_rotary_pos_emb,
        ),
    }
)

__ALL__ = [
    "ApproximationFunction",
    "ApproximationMixin",
    "NoApproximation",
    "TorchFunctionApproximation",
    "Approximate",
    "Approximator",
]


class ApproximationFunction:
    r"""
    This is an abstract class of approximation algorithm.
    Child classes to implement `execute()` and `from_shorthand()` method.
    """

    def __init__(self):
        pass

    def __str__(self) -> str:
        raise NotImplementedError

    def execute(self, *args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def from_shorthand(sh: str):
        if sh.startswith("NONE"):
            return NoApproximation.from_shorthand(sh)
        elif sh.startswith((*torch_function_mapping.keys(),)):
            return TorchFunctionApproximation.from_shorthand(sh)
        elif sh.startswith((*custom_function_mapping.keys(),)):
            return CustomFunctionApproximation.from_shorthand(sh)
        else:
            raise ValueError(f"unrecognized approximation function shorthand: {sh}")


class NoApproximation(ApproximationFunction):
    r"""
    This is a dummy approximation algorithm that means no approximation.
    """

    def __init__(self):
        super().__init__()

    def execute(self, *args, **kwargs):
        raise RuntimeError("NoApproximation is not supposed to be executed")

    @classmethod
    def from_shorthand(cls, sh: str):
        return cls()

    def __str__(self) -> str:
        return "Dummy approximation function: no approximation"

    def __repr__(self) -> str:
        return "NONE"


Identity = NoApproximation  # an alias, to be deprecated


class TorchFunctionApproximation(ApproximationFunction):
    r"""
    This class specifies an approximation function for a member of torch.nn.functional.
    """

    def __init__(self, func_id: None, algorithm: str = "vsimd", **extra_params):
        super().__init__()
        self.func_id = func_id
        self.torch_functional = torch_function_mapping[func_id]
        self.func_name = self.torch_functional.__name__
        self.algorithm = algorithm
        self.extra_params = extra_params

    @classmethod
    def from_shorthand(cls, sh: str):
        from dmx.compressor.utils.io import string_to_kwargs

        could_be_empty_str = lambda _s: _s
        could_be_empty_str.pattern = r".*"
        conf = parse(
            "{func_id:w}[{algorithm:w}]({extra_params:z})",
            sh,
            {"z": could_be_empty_str},
        )
        _func_id = conf["func_id"]
        _algo = conf["algorithm"]
        _extra_params = string_to_kwargs(conf["extra_params"])
        return cls(func_id=_func_id, algorithm=_algo, **_extra_params)

    def execute(self, *args, **kwargs):
        if self.algorithm == "vsimd":
            assert VSIMD_OP_REF_AVAILABLE, "SIMD op reference not available"
            return eval(f"vsimd.{self.func_name}")(*args, **kwargs, **self.extra_params)
        elif self.algorithm in ["experimental"]:
            return eval(f"{self.algorithm}.{self.func_name}")(
                *args, **kwargs, **self.extra_params
            )
        else:
            raise ValueError(
                f"unknown approximation algorithm {self.algorithm} for {self.func_id}"
            )

    def __str__(self) -> str:
        return f"Approximated version of {self.torch_functional} with annotation: {self.__repr__()}"

    def __repr__(self) -> str:
        from dmx.compressor.utils.io import kwargs_to_string

        return (
            f"{self.func_id}[{self.algorithm}]({kwargs_to_string(**self.extra_params)})"
        )


class CustomFunctionApproximation(ApproximationFunction):
    r"""
    This class specifies an approximation function for a custom written torch function.
    """

    def __init__(
        self,
        func_id: None,
        algorithm: str = "vsimd",
        **extra_params,
    ):
        super().__init__()
        self.func_id = func_id
        self.func_name, self.custom_functional = custom_function_mapping[func_id]
        self.algorithm = algorithm
        self.extra_params = extra_params

    @classmethod
    def from_shorthand(cls, sh: str):
        from dmx.compressor.utils.io import string_to_kwargs

        could_be_empty_str = lambda _s: _s
        could_be_empty_str.pattern = r".*"
        conf = parse(
            "{func_id:w}[{algorithm:w}]({extra_params:z})",
            sh,
            {"z": could_be_empty_str},
        )
        _func_id = conf["func_id"]
        _algo = conf["algorithm"]
        _extra_params = string_to_kwargs(conf["extra_params"])
        return cls(func_id=_func_id, algorithm=_algo, **_extra_params)

    def execute(self, *args, **kwargs):
        if self.algorithm == "vsimd":
            assert VSIMD_OP_REF_AVAILABLE, "SIMD op reference not available"
            return eval(f"vsimd.{self.func_name}")(*args, **kwargs, **self.extra_params)
        elif self.algorithm in ["experimental"]:
            return eval(f"{self.algorithm}.{self.func_name}")(
                *args, **kwargs, **self.extra_params
            )
        else:
            raise ValueError(
                f"unknown approximation algorithm {self.algorithm} for {self.func_id}"
            )

    def __str__(self) -> str:
        return f"Approximated version of {self.custom_functional} with annotation: {self.__repr__()}"

    def __repr__(self) -> str:
        from dmx.compressor.utils.io import kwargs_to_string

        return (
            f"{self.func_id}[{self.algorithm}]({kwargs_to_string(**self.extra_params)})"
        )


class Approximate(nn.Module):
    r"""
    An approximation operator container
    """

    def __init__(self, function=NoApproximation()):
        super().__init__()
        self.set_function(function)

    def set_function(self, function: Union[str, ApproximationFunction]) -> None:
        if not isinstance(function, ApproximationFunction):
            function = ApproximationFunction.from_shorthand(function)
        self.function = function

    def forward(self, input, *args, **kwargs):
        return self.function.execute(input, *args, **kwargs)

    def extra_repr(self):
        return f"function = {self.function.__repr__()}"


class Approximator(nn.Module):
    r"""
    Encapsulates approximation forward logic of a module
    """

    def __init__(self, function=Identity()):
        super().__init__()
        if not isinstance(function, ApproximationFunction):
            function = ApproximationFunction.from_shorthand(function)
        self.function = function

    def forward(self, input):
        _output = self.function.execute(input)[0]
        if not isinstance(
            self.function,
            (NoApproximation,),
        ):
            with torch.no_grad():
                self.approximation_error = _output - input
        return _output

    def extra_repr(self):
        return f"function = {self.function.__repr__()}"

    def approximation_function(self):
        return repr(self.function)


class ApproximationMixin:
    r"""
    Mixin to equip modules with approximated forward logic through Approximator
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.approximator = Approximate()  # if isinstance(self, DmxModule) else None
        self.approximation_error = None

    def approx_forward(self, inputs, *args, **kwargs):
        if not self.functional_forward is None:
            _output = self.functional_forward(*inputs, *args, **kwargs)
        else:
            _output = super().forward(*inputs)
        if not isinstance(
            self.approximator.function,
            (NoApproximation,),
        ):
            with torch.no_grad():
                _approx = self.approximator(*inputs, *args, **kwargs)
                if isinstance(_approx,tuple): #for modules that return multiple values
                    assert isinstance(_output,tuple), \
                        'module and its approximation should both return a tuple'
                    self.approximation_error = [x - y.data for x,y in zip(_approx,_output)]
                    for x,y in zip(_approx,_output):
                        y.data = x
                else:
                    self.approximation_error = _approx - _output.data
                    _output.data = _approx
        return _output

    @property
    def approximation_function(self):
        return self.approximator.function



import torch
import os
from pathlib import Path
import hashlib
import tempfile
import shutil
import yaml
from typing import Type

from dmx.compressor.modeling.nn import DmxModule, DmxModuleType
from dmx.compressor.numerical import Format
from dmx.compressor.sparse import Sparseness
from dmx.compressor.functional import ApproximationFunction


def compute_md5(file_name):
    hash_md5 = hashlib.md5()
    with open(file_name, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def save_state_dict_and_return_url(module: torch.nn.Module, parent_dir: str) -> str:
    temp_file_name = tempfile.NamedTemporaryFile().name
    torch.save(module.state_dict(), temp_file_name)
    md5 = compute_md5(temp_file_name)
    file_name = os.path.join(parent_dir, f"{md5}.pth")
    shutil.move(temp_file_name, file_name)
    return Path(os.path.abspath(file_name)).as_uri()


def dmx_module_instance_constructor(
    loader: yaml.SafeLoader, node: yaml.nodes.ScalarNode
) -> str:
    return eval(f"nn.{node.value}")


def dmx_module_instance_representer(
    dumper: yaml.SafeDumper, val: Type[DmxModule]
) -> yaml.nodes.ScalarNode:
    return dumper.represent_scalar("!DmxModule", val.__name__, style=None)


def format_constructor(loader: yaml.SafeLoader, node: yaml.nodes.ScalarNode) -> str:
    return Format.from_shorthand(node.value)


def format_representer(dumper: yaml.SafeDumper, val: Format) -> yaml.nodes.ScalarNode:
    return dumper.represent_scalar("!Format", repr(val), style=None)


def sparseness_constructor(loader: yaml.SafeLoader, node: yaml.nodes.ScalarNode) -> str:
    return Sparseness.from_shorthand(node.value)


def sparseness_representer(
    dumper: yaml.SafeDumper, val: Format
) -> yaml.nodes.ScalarNode:
    return dumper.represent_scalar("!Sparseness", repr(val), style=None)


def approximation_function_constructor(
    loader: yaml.SafeLoader, node: yaml.nodes.ScalarNode
) -> str:
    return ApproximationFunction.from_shorthand(node.value)


def approximation_function_representer(
    dumper: yaml.SafeDumper, val: Format
) -> yaml.nodes.ScalarNode:
    return dumper.represent_scalar("!ApproximationFunction", repr(val), style=None)


def get_loader():
    loader = yaml.SafeLoader
    loader.add_constructor("!DmxModule", dmx_module_instance_constructor)
    loader.add_constructor("!Format", format_constructor)
    loader.add_constructor("!Sparseness", sparseness_constructor)
    loader.add_constructor("!ApproximationFunction", approximation_function_constructor)
    return loader


def get_dumper():
    dumper = yaml.SafeDumper
    dumper.add_multi_representer(DmxModuleType, dmx_module_instance_representer)
    dumper.add_multi_representer(Format, format_representer)
    dumper.add_multi_representer(Sparseness, sparseness_representer)
    dumper.add_multi_representer(
        ApproximationFunction, approximation_function_representer
    )
    dumper.ignore_aliases = lambda _self, _data: True
    return dumper


def load_config_file(config_file="configs/dmx.yaml"):
    with open(config_file, "r") as f:
        config = yaml.load(f, Loader=get_loader())
    return config


def save_config_file(config, config_file="configs/dmx.yaml"):
    with open(config_file, "w") as f:
        f.write(
            yaml.dump(
                config,
                Dumper=get_dumper(),
            )
        )


def string_to_kwargs(kwargs_string):
    """Parses a string of keyword arguments into a dictionary.

    Args:
        kwargs_string: A string of keyword arguments in the format "key1=value1, key2=value2, ...".

    Returns:
        A dictionary containing the parsed keyword arguments.
    """
    kwargs = {}
    if kwargs_string:
        for item in kwargs_string.split(","):
            key, value = item.split("=")
            value = value.strip()
            try:
                _value = eval(value)
            except:
                _value = value 
            kwargs[key.strip()] = _value
    return kwargs


def kwargs_to_string(**kwargs):
    """Converts kwargs dictionary to a comma-separated string of key=value pairs."""
    return ", ".join(f"{key}={value}" for key, value in kwargs.items())



try:
    from .quant_function import *
except ImportError as error:
    print("Error importing Block Quantize CUDA kernels")

__all__ = [
    "fixed_point_quantize",
    "block_quantize",
    "float_quantize",
]



from types import SimpleNamespace
from dmx.compressor.modeling.nn.core import DmxModule
import torch
from functools import partialmethod

torch.nn.Module.load_state_dict = partialmethod(
    torch.nn.Module.load_state_dict, strict=False
)

from .numerical import Format, NUMERICS_UTILS_AVAILABLE
from .sparse import Sparseness
from .functional import ApproximationFunction, VSIMD_OP_REF_AVAILABLE
from .modeling import (
    nn,
    DmxModel,
    DmxConfigRule,
)

# Numerical format aliases
format = SimpleNamespace(
    SAME=Format.from_shorthand("SAME"),
    FLOAT32=Format.from_shorthand("FP[1|8|23,127](_N)"),
    FLOAT16=Format.from_shorthand("FP[1|5|10,15](FN)"),
    BFLOAT16=Format.from_shorthand("FP[1|8|7,127](FN)"),
    AFLOAT8=Format.from_shorthand("FP[1|4|3,7](_N)"),
    BFLOAT8=Format.from_shorthand("FP[1|5|2,15](_N)"),
    INT8=Format.from_shorthand("XP[8,0](CSN)"),
    INT4=Format.from_shorthand("XP[4,0](CSN)"),
    BFP32_1=Format.from_shorthand("BFP[24|8]{1}(SN)"),
    BFP24_64=Format.from_shorthand("BFP[16|8]{64}(SN)"),
    BFP24_32=Format.from_shorthand("BFP[16|8]{32}(SN)"),
    BFP24_16=Format.from_shorthand("BFP[16|8]{16}(SN)"),
    BFP16_128=Format.from_shorthand("BFP[8|8]{128}(SN)"),
    BFP16_64=Format.from_shorthand("BFP[8|8]{64}(SN)"),
    BFP16_32=Format.from_shorthand("BFP[8|8]{32}(SN)"),
    BFP16_16=Format.from_shorthand("BFP[8|8]{16}(SN)"),
    BFP14_128=Format.from_shorthand("BFP[6|8]{128}(SN)"),
    BFP14_64=Format.from_shorthand("BFP[6|8]{64}(SN)"),
    BFP14_32=Format.from_shorthand("BFP[6|8]{32}(SN)"),
    BFP14_16=Format.from_shorthand("BFP[6|8]{16}(SN)"),
    BFP12_128=Format.from_shorthand("BFP[4|8]{128}(SN)"),
    BFP12_64=Format.from_shorthand("BFP[4|8]{64}(SN)"),
    BFP12_32=Format.from_shorthand("BFP[4|8]{32}(SN)"),
    BFP12_16=Format.from_shorthand("BFP[4|8]{16}(SN)"),
    BFP16A_128=Format.from_shorthand("BFP[8|8]{128}(_N)"),
    BFP16A_64=Format.from_shorthand("BFP[8|8]{64}(_N)"),
    BFP16A_32=Format.from_shorthand("BFP[8|8]{32}(_N)"),
    BFP16A_16=Format.from_shorthand("BFP[6|8]{16}(_N)"),
    BFP14A_128=Format.from_shorthand("BFP[6|8]{128}(_N)"),
    BFP14A_64=Format.from_shorthand("BFP[6|8]{64}(_N)"),
    BFP14A_32=Format.from_shorthand("BFP[6|8]{32}(_N)"),
    BFP14A_16=Format.from_shorthand("BFP[6|8]{16}(_N)"),
    BFP12A_128=Format.from_shorthand("BFP[4|8]{128}(_N)"),
    BFP12A_64=Format.from_shorthand("BFP[4|8]{64}(_N)"),
    BFP12A_32=Format.from_shorthand("BFP[4|8]{32}(_N)"),
    BFP12A_16=Format.from_shorthand("BFP[4|8]{16}(_N)"),
    SBFP12_16=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,7](FN)>{16}"),
    SBFP12_16_4=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,4](FN)>{16}"),
    SBFP12_16_5=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,5](FN)>{16}"),
    SBFP12_16_6=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,6](FN)>{16}"),
    SBFP12_16_7=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,7](FN)>{16}"),
    SBFP12_16_8=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,8](FN)>{16}"),
    SBFP12_16_9=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,9](FN)>{16}"),
    SBFP12_16_10=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,10](FN)>{16}"),
    SBFP12_16_11=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,11](FN)>{16}"),
    SBFP12_16_12=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,12](FN)>{16}"),
    SBFP12_16_13=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,13](FN)>{16}"),
    SBFP12_16_14=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,14](FN)>{16}"),
    SBFP12_16_15=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,15](FN)>{16}"),
    SBFP12_16_16=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,16](FN)>{16}"),
    SBFP12_16_17=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,17](FN)>{16}"),
    SBFP12_16_18=Format.from_shorthand("SBFP<XP[4,0](CSN)><FP[0|4|4,18](FN)>{16}"),
    MXFP8_E4M3K128=Format.from_shorthand("MXFP8[E4M3]{128}"),
    MXFP8_E4M3K64=Format.from_shorthand("MXFP8[E4M3]{64}"),
    MXFP8_E4M3K32=Format.from_shorthand("MXFP8[E4M3]{32}"),
    MXFP8_E5M2K128=Format.from_shorthand("MXFP8[E5M2]{128}"),
    MXFP8_E5M2K64=Format.from_shorthand("MXFP8[E5M2]{64}"),
    MXFP8_E5M2K32=Format.from_shorthand("MXFP8[E5M2]{32}"),
    MXFP6_E2M3K128=Format.from_shorthand("MXFP6[E2M3]{128}"),
    MXFP6_E2M3K64=Format.from_shorthand("MXFP6[E2M3]{64}"),
    MXFP6_E2M3K32=Format.from_shorthand("MXFP6[E2M3]{32}"),
    MXFP6_E3M2K128=Format.from_shorthand("MXFP6[E3M2]{128}"),
    MXFP6_E3M2K64=Format.from_shorthand("MXFP6[E3M2]{64}"),
    MXFP6_E3M2K32=Format.from_shorthand("MXFP6[E3M2]{32}"),
    MXFP4_E2M1K128=Format.from_shorthand("MXFP4[E2M1]{128}"),
    MXFP4_E2M1K64=Format.from_shorthand("MXFP4[E2M1]{64}"),
    MXFP4_E2M1K32=Format.from_shorthand("MXFP4[E2M1]{32}"),
    MXINT8_K128=Format.from_shorthand("MXINT8{128}"),
    MXINT8_K64=Format.from_shorthand("MXINT8{64}"),
    MXINT8_K32=Format.from_shorthand("MXINT8{32}"),
    MXINT6_K128=Format.from_shorthand("MXINT6{128}"),
    MXINT6_K64=Format.from_shorthand("MXINT6{64}"),
    MXINT6_K32=Format.from_shorthand("MXINT6{32}"),
    MXINT4_K128=Format.from_shorthand("MXINT4{128}"),
    MXINT4_K64=Format.from_shorthand("MXINT4{64}"),
    MXINT4_K32=Format.from_shorthand("MXINT4{32}"),
)

# Sparseness aliases
sparseness = SimpleNamespace(
    BTK8_4_LD=Sparseness.from_shorthand("BTOPK{4:8,-1}(U)"),
    BTK8_4_FD=Sparseness.from_shorthand("BTOPK{4:8,1}(U)"),
    BTK8_2_LD=Sparseness.from_shorthand("BTOPK{2:8,-1}(U)"),
    BTK8_2_FD=Sparseness.from_shorthand("BTOPK{2:8,1}(U)"),
)

# Default approximation function aliases
default_approx = SimpleNamespace(
    RELU=ApproximationFunction.from_shorthand("NONE"),
    RELU6=ApproximationFunction.from_shorthand("NONE"),
    SILU=ApproximationFunction.from_shorthand(
        "SILU[vsimd]()" if VSIMD_OP_REF_AVAILABLE else "NONE"
    ),
    SOFTMAX=ApproximationFunction.from_shorthand(
        "SOFTMAX[vsimd](max_adjust=0.1141)" if VSIMD_OP_REF_AVAILABLE else "NONE"
    ),
    GELU=ApproximationFunction.from_shorthand("NONE"),
    QUICK_GELU=ApproximationFunction.from_shorthand(
        "QUICK_GELU[vsimd]()" if VSIMD_OP_REF_AVAILABLE else "NONE"
    ),
    TANH=ApproximationFunction.from_shorthand("NONE"),
    BATCH_NORM_2D=ApproximationFunction.from_shorthand("NONE"),
    LAYER_NORM=ApproximationFunction.from_shorthand(
        "LAYER_NORM[vsimd]()" if VSIMD_OP_REF_AVAILABLE else "NONE"
    ),
    RMS_NORM=ApproximationFunction.from_shorthand(
        "RMS_NORM[vsimd]()" if VSIMD_OP_REF_AVAILABLE else "NONE"
    ),
    GROUP_NORM=ApproximationFunction.from_shorthand("NONE"),
    EXP=ApproximationFunction.from_shorthand(
        "EXP[vsimd](knorm=0,kmax=15,use_exp_large=True)"
        if VSIMD_OP_REF_AVAILABLE
        else "NONE"
    ),
    APPLY_LLAMA_ROPE=ApproximationFunction.from_shorthand(
        "APPLY_LLAMA_ROPE[vsimd]()" if VSIMD_OP_REF_AVAILABLE else "NONE"
    ),
    NONE=ApproximationFunction.from_shorthand("NONE"),
)

# Automatic configuration rules
config_rules = SimpleNamespace(
    BASELINE=[
        DmxConfigRule(
            module_types=(nn.Linear,),
            module_config=dict(
                input_formats=[format.SAME],
                weight_format=format.SAME,
                bias_format=format.SAME,
                output_formats=[format.SAME],
            ),
        ),
        DmxConfigRule(
            module_types=(
                nn.Conv1d,
                nn.Conv2d,
                nn.ConvTranspose2d,
            ),
            module_config=dict(
                input_formats=[format.SAME],
                weight_format=format.SAME,
                bias_format=format.SAME,
                output_formats=[format.SAME],
            ),
        ),
        DmxConfigRule(
            module_types=(nn.ResAdd,),
            module_config=dict(
                input_formats=[format.SAME, format.SAME],
                output_formats=[format.SAME],
            ),
        ),
        DmxConfigRule(
            module_types=(nn.ActActMatMul,),
            module_config=dict(
                input_formats=[format.SAME, format.SAME],
                output_formats=[format.SAME],
            ),
        ),
        DmxConfigRule(
            module_types=(nn.Embedding,),
            module_config=dict(
                output_formats=[format.SAME],
            ),
        ),
        DmxConfigRule(
            module_types=(
                nn.MaxPool2d,
                nn.AdaptiveAvgPool2d,
                nn.AvgPool2d,
            ),
            module_config=dict(
                input_formats=[format.SAME],
                output_formats=[format.SAME],
            ),
        ),
        DmxConfigRule(
            module_types=(
                nn.ReLU,
                nn.ReLU6,
                nn.GELUBase,
                nn.SiLU,
                nn.Tanh,
                nn.Softmax,
                nn.LayerNorm,
                nn.BatchNorm2d,
                nn.GroupNorm,
                nn.Exp,
            ),
            module_config=dict(
                input_formats=[format.SAME],
                output_formats=[format.SAME],
                approximation_function=default_approx.NONE,
            ),
        ),
    ],
    BASIC=[
        DmxConfigRule(
            module_types=(nn.Linear,),
            module_config=dict(
                input_formats=[format.BFP16_64],
                weight_format=format.BFP16_64,
                bias_format=format.BFP32_1,
                output_formats=[format.FLOAT16],
            ),
        ),
        DmxConfigRule(
            module_types=(
                nn.Conv1d,
                nn.Conv2d,
                nn.ConvTranspose2d,
            ),
            module_config=dict(
                input_formats=[format.BFP16_64],
                weight_format=format.BFP16_64,
                bias_format=format.BFP32_1,
                output_formats=[format.FLOAT16],
            ),
        ),
        DmxConfigRule(
            module_types=(nn.ResAdd,),
            module_config=dict(
                input_formats=[format.FLOAT16, format.FLOAT16],
                output_formats=[format.FLOAT16],
            ),
        ),
        DmxConfigRule(
            module_types=(nn.ActActMatMul,),
            module_config=dict(
                input_formats=[format.BFP16_64, format.BFP16_64],
                output_formats=[format.FLOAT16],
            ),
        ),
        DmxConfigRule(
            module_types=(nn.Embedding,),
            module_config=dict(
                output_formats=[format.FLOAT16],
            ),
        ),
        DmxConfigRule(
            module_types=(
                nn.MaxPool2d,
                nn.AdaptiveAvgPool2d,
                nn.AvgPool2d,
            ),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
            ),
        ),
        DmxConfigRule(
            module_types=(nn.ReLU,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.RELU,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.ReLU6,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.RELU6,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.GELUBase,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.GELU,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.QuickGELU,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.QUICK_GELU,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.SiLU,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.SILU,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.Tanh,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.TANH,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.Softmax,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.SOFTMAX,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.LayerNorm,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.LAYER_NORM,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.RMSNorm,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.RMS_NORM,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.BatchNorm2d,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.BATCH_NORM_2D,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.GroupNorm,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.GROUP_NORM,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.Exp,),
            module_config=dict(
                input_formats=[format.FLOAT16],
                output_formats=[format.FLOAT16],
                approximation_function=default_approx.EXP,
            ),
        ),
        DmxConfigRule(
            module_types=(nn.ApplyRotaryPosEmb,),
            module_config=dict(
                input_formats=[
                    format.FLOAT16,
                    format.FLOAT16,
                    format.FLOAT16,
                    format.FLOAT16,
                ],
                output_formats=[format.FLOAT16, format.FLOAT16],
                approximation_function=default_approx.APPLY_LLAMA_ROPE,
            ),
        ),
    ],
    SBFP_WEIGHT_STORAGE=[
        DmxConfigRule(
            module_types=(
                nn.Linear,
                nn.Conv1d,
                nn.Conv2d,
                nn.ConvTranspose2d,
            ),
            module_config=dict(
                weight_storage_format=format.SBFP12_16,
            ),
        ),
    ],
)



#!/usr/bin/env python3

import pytest
import torch
from torch import Tensor
import torch.nn as nn
from dmx.compressor.modeling import nn as dmxnn
from dmx.compressor.fx.tracer import QuantTracer
from dmx.compressor.fx.transform import cast_input_output_transform
import torch.fx as fx
from transformers.pytorch_utils import Conv1D

RANDOM_SEED = 0

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

torch.manual_seed(RANDOM_SEED)


class LeNet(nn.Module):
    def __init__(self, hidden_dims, input_dim=784, output_dim=10) -> None:
        super(LeNet, self).__init__()
        self.input_layer = nn.Linear(input_dim, hidden_dims[0])
        self.intermediate_layers = nn.ModuleList(
            [
                nn.Linear(hidden_dims[i], hidden_dims[i + 1])
                for i in range(len(hidden_dims) - 1)
            ]
        )
        self.output_layer = nn.Linear(hidden_dims[-1], output_dim)
        self.act_func = nn.ReLU(inplace=True)

    def forward(self, x: Tensor) -> Tensor:
        x = self.input_layer(x)
        x = self.act_func(x)
        for layer in self.intermediate_layers:
            x = layer(x)
            x = self.act_func(x)
        x = self.output_layer(x)
        return x


def checkTransform(gm: nn.Module) -> bool:
    if not isinstance(gm, fx.graph_module.GraphModule):
        return True
    nodeList = []
    for i in gm.graph.nodes:
        nodeList.append(i)
    for i in nodeList:
        if i.op == "placeholder":
            if i.target + "_cast" not in i.next.target:
                return False
        elif i.op == "output":
            if "output_cast" not in i.prev.target:
                return False
        elif i.op == "get_attr" and "weight" in i.target:
            if not (
                "weight_cast" in i.next.target
                or (
                    ("sparsifier" in i.next.target or "approximator" in i.next.target)
                    and "weight_cast" in i.next.next.target
                )
                or (
                    "sparsifier" in i.next.target
                    and "approximator" in i.next.next.target
                    and "weight_cast" in i.next.next.next.target
                )
            ):
                return False
        elif i.op == "get_attr" and "bias" in i.target:
            if "weight_cast" not in i.next.target and "bias_cast" not in i.next.target:
                return False
        elif i.op == "call_module":
            if not checkTransform(gm.get_submodule(i.target)):
                return False
    return True


@pytest.mark.skip()
def test_dmx_transform():
    net = Net()
    cnet = DmxNet()
    gm_before = fx.symbolic_trace(net)
    assert not checkTransform(gm_before), True
    gm = cast_input_output_transform(net, QuantTracer())
    input = torch.rand(1, 64)

    # Initialize bias and weight
    weight = torch.rand(64, 64)
    bias = torch.zeros(1, 64)
    cnet.linear.bias.data = bias
    gm.linear.bias.data = bias
    cnet.linear.weight.data = weight
    gm.linear.weight.data = weight
    coutput = cnet(input)
    output = gm(input)
    assert (coutput - output).abs().sum() == 0, True
    assert checkTransform(gm), True


def test_conv1D():
    net = Conv1D(32, 32)
    gm = cast_input_output_transform(net, QuantTracer())


def test_corsiar_transform_sparsify_approximate():
    net = Net()
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


@pytest.mark.skip()
def test_double_transform():
    net = Net()
    with pytest.raises(Warning):
        gm = cast_input_output_transform(net, QuantTracer())
        gm2t = cast_input_output_transform(gm)


class downcast(nn.Module):
    def __init__(self, format="SAME", dump_to=None):
        super().__init__()

    def forward(self, x):
        return x


class upcast(nn.Module):
    def __init__(self, format="SAME", dump_to=None):
        super().__init__()

    def forward(self, x):
        return x


class weightCast(nn.Module):
    def __init__(self, format="SAME", dump_to=None):
        super().__init__()

    def forward(self, x):
        return x


def test_fakecast_transform():
    net = Net()
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


@pytest.mark.parametrize(
    "layers",
    (
        [10, 10],
        [10, 10, 10, 10, 10, 10],
    ),
)
def test_lenet_1hid_dmx_transform_without_cfg(layers):
    net = LeNet(layers)
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


@pytest.mark.parametrize(
    "layers",
    ([10, 10],),
)
def test_lenet_1hid_dmx_transform_with_test_cfg(layers):
    net = LeNet(layers)
    gm = cast_input_output_transform(
        net, QuantTracer(), cfg="./tests/configs/lenet_test.yaml"
    )
    assert checkTransform(gm), True


@pytest.mark.parametrize(
    "layers",
    (
        [10, 10],
        [10, 10, 10, 10, 10, 10],
    ),
)
def test_lenet_fakecast_transform(layers):
    net = LeNet(layers)
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_conv2D_fakecast_transform():
    net = torch.nn.Conv2d(16, 8, 2)
    cnet = dmxnn.Conv2d(16, 8, 2)
    input = torch.rand(1, 16, 16, 16)
    weight = torch.rand(8, 16, 2, 2)
    bias = torch.rand(8)
    net.weight.data = weight
    cnet.weight.data = weight
    net.bias.data = bias
    cnet.bias.data = bias
    coutput = cnet(input)

    gm = cast_input_output_transform(net, QuantTracer())
    output = gm(input)
    assert (coutput - output).abs().sum() < 1e-5, True
    assert checkTransform(gm), True


def test_Dropout_transform():
    net = torch.nn.Dropout()
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_AdaptiveAvgPool2d_transform():
    net = torch.nn.AdaptiveAvgPool2d(16)
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_AvgPool2d_transform():
    net = torch.nn.AvgPool2d(8)
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_MaxPool2d_transform():
    net = torch.nn.MaxPool2d(16)
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_Softmax_transform():
    net = torch.nn.Softmax()
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_LayerNorm_transform():
    net = torch.nn.LayerNorm((1, 64))
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_ReLU_transform():
    net = torch.nn.ReLU()
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm)


def test_ReLU6_transform():
    net = torch.nn.ReLU6()
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_Tanh_transform():
    net = torch.nn.Tanh()
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


def test_GELU_transform():
    net = torch.nn.GELU()
    gm = cast_input_output_transform(net, QuantTracer())
    assert checkTransform(gm), True


class Net(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.linear = nn.Linear(64, 64)

    def forward(self, input):
        return self.linear(input)


class DmxNet(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.linear = dmxnn.Linear(64, 64)

    def forward(self, input):
        return self.linear(input)



import pytest
import torch
from typing import Tuple, Optional, Union

from dmx.compressor import format, sparseness, sparse, numerical
from dmx.compressor import nn as dmxnn
from dataclasses import dataclass

RANDOM_SEED = 0
torch.manual_seed(RANDOM_SEED)


@dataclass(frozen=True)
class TestCase:
    # Given:
    layer_type: dmxnn.DmxModule
    layer_dims: Tuple[
        int, int, Optional[Tuple[int, ...]], Optional[int], Optional[int], Optional[str]
    ]
    input_dims: Tuple[int, Optional[Tuple[int, ...]]]
    weight_sparseness: sparse.Sparseness
    input_format: Union[numerical.Format, torch.dtype]
    weight_format: Union[numerical.Format, torch.dtype]
    # Ensure:
    _weight_elem_count: float
    _weight_size_in_bytes: float
    _flops: float
    _bops: float

    def __str__(self):
        return f"{self.layer_type.__name__}-{self.layer_dims}-{self.input_dims}-{repr(self.weight_sparseness)}-{repr(self.input_format)}-{repr(self.weight_format)}"


test_cases = [
    TestCase(
        layer_type=dmxnn.Linear,
        layer_dims=(16, 16, None, None, None, None),
        input_dims=(1, None),
        weight_sparseness=sparse.Dense(),
        input_format=torch.float32,
        weight_format=torch.float32,
        # --------------------------------
        _weight_elem_count=256.0,
        _weight_size_in_bytes=1024.0,
        _flops=256.0,
        _bops=262144.0,
    ),
    TestCase(
        layer_type=dmxnn.Linear,
        layer_dims=(8, 24, None, None, None, None),
        input_dims=(8, None),
        weight_sparseness=sparseness.BTK8_4_LD,
        input_format=format.BFP16_128,
        weight_format=format.BFP12_128,
        # --------------------------------
        _weight_elem_count=96.0,
        _weight_size_in_bytes=48.75,
        _flops=768.0,
        _bops=25155.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv2d,
        layer_dims=(8, 32, (3, 3), 1, 1, "same"),
        input_dims=(8, (10, 10)),
        weight_sparseness=sparse.Dense(),
        input_format=torch.float16,
        weight_format=torch.float16,
        # --------------------------------
        _weight_elem_count=2304.0,
        _weight_size_in_bytes=4608.0,
        _flops=1843200.0,
        _bops=471859200.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv2d,
        layer_dims=(8, 32, (3, 3), 1, 1, "valid"),
        input_dims=(8, (10, 10)),
        weight_sparseness=sparse.Dense(),
        input_format=torch.float16,
        weight_format=torch.float16,
        # --------------------------------
        _weight_elem_count=2304.0,
        _weight_size_in_bytes=4608.0,
        _flops=1179648.0,
        _bops=301989888.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv2d,
        layer_dims=(8, 32, (3, 3), 1, 1, "same"),
        input_dims=(8, (10, 10)),
        weight_sparseness=sparseness.BTK8_4_FD,
        input_format=torch.float16,
        weight_format=torch.float16,
        # --------------------------------
        _weight_elem_count=1152.0,
        _weight_size_in_bytes=2304.0,
        _flops=921600.0,
        _bops=235929600.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv2d,
        layer_dims=(8, 32, (5, 5), 1, 1, "valid"),
        input_dims=(8, (10, 10)),
        weight_sparseness=sparseness.BTK8_2_FD,
        input_format=format.INT8,
        weight_format=format.INT8,
        # --------------------------------
        _weight_elem_count=1600.0,
        _weight_size_in_bytes=1600.0,
        _flops=460800.0,
        _bops=29491200.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv2d,
        layer_dims=(8, 32, (7, 7), 1, 1, "same"),
        input_dims=(16, (13, 13)),
        weight_sparseness=sparseness.BTK8_2_FD,
        input_format=format.INT8,
        weight_format=format.INT4,
        # --------------------------------
        _weight_elem_count=3136.0,
        _weight_size_in_bytes=1568.0,
        _flops=8479744.0,
        _bops=271351808.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv2d,
        layer_dims=(8, 32, (7, 7), 1, 1, "valid"),
        input_dims=(16, (13, 13)),
        weight_sparseness=sparseness.BTK8_2_FD,
        input_format=format.INT8,
        weight_format=format.INT4,
        # --------------------------------
        _weight_elem_count=3136.0,
        _weight_size_in_bytes=1568.0,
        _flops=2458624.0,
        _bops=78675968.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv2d,
        layer_dims=(8, 32, (7, 7), 1, 4, "same"),
        input_dims=(8, (13, 13)),
        weight_sparseness=sparse.Dense(),
        input_format=format.INT4,
        weight_format=format.INT4,
        # --------------------------------
        _weight_elem_count=3136.0,
        _weight_size_in_bytes=1568.0,
        _flops=4239872.0,
        _bops=67837952.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv2d,
        layer_dims=(32, 32, (3, 3), 2, 4, "valid"),
        input_dims=(8, (10, 10)),
        weight_sparseness=sparseness.BTK8_4_FD,
        input_format=format.BFP14_128,
        weight_format=format.BFP12_128,
        # --------------------------------
        _weight_elem_count=1152.0,
        _weight_size_in_bytes=585.0,
        _flops=147456.0,
        _bops=3631680.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv1d,
        layer_dims=(8, 16, (3,), 1, 2, "same"),
        input_dims=(16, (10,)),
        weight_sparseness=sparse.Dense(),
        input_format=torch.float16,
        weight_format=torch.float16,
        # --------------------------------
        _weight_elem_count=192.0,
        _weight_size_in_bytes=384.0,
        _flops=30720.0,
        _bops=7864320.0,
    ),
    TestCase(
        layer_type=dmxnn.Conv1d,
        layer_dims=(8, 16, (3,), 2, 2, "valid"),
        input_dims=(16, (10,)),
        weight_sparseness=sparse.Dense(),
        input_format=format.INT8,
        weight_format=format.INT8,
        # --------------------------------
        _weight_elem_count=192.0,
        _weight_size_in_bytes=192.0,
        _flops=12288.0,
        _bops=786432.0,
    ),
]


def _create_module(cls, layer_dims, weight_sparseness, input_format, weight_format):
    ch_in, ch_out, ker_size, stride, groups, padding = layer_dims
    if cls == dmxnn.Linear:
        _module = cls(
            in_features=ch_in,
            out_features=ch_out,
        )
    elif cls in (dmxnn.Conv1d, dmxnn.Conv2d):
        _module = cls(
            in_channels=ch_in,
            out_channels=ch_out,
            kernel_size=ker_size,
            stride=stride,
            groups=groups,
            padding=padding,
        )
    else:
        raise ValueError("unsupported module class {cls}")
    _module.weight_sparsifier.configure(sparseness=weight_sparseness)
    if isinstance(input_format, numerical.Format):
        _module.input_casts.input_cast.set_format(input_format)
    if isinstance(weight_format, numerical.Format):
        _module.weight_cast.set_format(weight_format)
    elif weight_format == torch.float16:  # in order to work on both CPU/GPU
        _module.weight_cast.set_format(format.FLOAT16)
    return _module


def _create_test_input(batch_size, in_channel, img_size, input_format):
    shape = (batch_size, in_channel)
    if img_size is not None:
        shape += img_size
    dtype = input_format if isinstance(input_format, torch.dtype) else torch.float32
    return torch.randn(torch.Size(shape)).to(dtype)


@pytest.fixture(
    scope="class",
    params=test_cases,
    ids=[str(tc) for tc in test_cases],
)
def _test_run(request):
    p = request.param
    ch_in, ch_out, ker_size, stride, groups, padding = p.layer_dims
    batch_size, img_size = p.input_dims
    module = _create_module(
        p.layer_type,
        p.layer_dims,
        p.weight_sparseness,
        p.input_format,
        p.weight_format,
    )
    x_ref = _create_test_input(
        batch_size=batch_size,
        in_channel=ch_in,
        img_size=img_size,
        input_format=p.input_format,
    )
    with module.counting_flops(zero=True), torch.no_grad():
        module(x_ref)

    yield module, p


@pytest.mark.usefixtures("_test_run")
def test_perf_proxy(_test_run):
    module, p = _test_run
    assert module.weight_elem_count == p._weight_elem_count
    assert module.weight_size_in_bytes == p._weight_size_in_bytes
    assert module.flops == p._flops
    assert module.bops == p._bops



from torch.fx import Interpreter
from .utils import dmx_aware_functional_mappings
from torch.fx.node import Node
from typing import Any
import torch


class RecordInputInterpreter(Interpreter):
    """
    Interpreter that captures the input tensors to the compound functions.
    attributes:
        nodeInputs: dictionary that maps node name to its tensor args and kwargs
    """

    def __init__(self, module, garbage_collect_values=True, graph=None):
        super().__init__(module, garbage_collect_values, graph)
        self.nodeInputs = {}

    def run_node(self, n: Node) -> Any:
        """
        Run a specific node ``n`` and return the result.
        Calls into placeholder, get_attr, call_function,
        call_method, call_module, or output depending
        on ``node.op``

        Record input tensors to nodeInputs if target is in compound_functions

        Args:
            n (Node): The Node to execute

        Returns:
            Any: The result of executing ``n``
        """
        with self._set_current_node(n):
            args, kwargs = self.fetch_args_kwargs_from_env(n)
            assert isinstance(args, tuple)
            assert isinstance(kwargs, dict)

            target = n.target

            # recording inputs
            if str(target) in dmx_aware_functional_mappings and (
                dmx_aware_functional_mappings[str(target)].is_compound
                or str(target)
                in (
                    "<built-in function mul>",
                    "<built-in function add>",
                )
            ):
                self.nodeInputs[n.name] = (args, kwargs)

            device = None
            if (
                target in self.submodules
                and hasattr(self.submodules[target], "weight")
                and self.submodules[target].weight is not None
            ):
                device = self.submodules[target].weight.device
            for a in args:
                if isinstance(a, torch.Tensor) and device is None:
                    device = a.device
            if device is not None:
                args = tuple(
                    a.to(device) if isinstance(a, torch.Tensor) else a for a in args
                )
                kwargs = {
                    k: v.to(device) if isinstance(v, torch.Tensor) else v
                    for k, v in kwargs.items()
                }
            return getattr(self, n.op)(n.target, args, kwargs)



#!/usr/bin/env python3
import copy

from typing import Any, Callable, Dict, Optional, Tuple, Union

import torch
import torch.fx as fx
from torch.fx.node import Argument, Node, Target, map_arg, map_aggregate
from torch.fx import Graph, Proxy
import torch.fx.traceback as fx_traceback

import itertools

from dmx.compressor.fx.transformer.utils import process_args

from .utils import dmx_aware_mapping, dmx_aware_functional_mappings


class QdQTransformer(fx.Transformer):
    def __init__(self, module: fx.GraphModule, scopeDict: dict = None, cfg=None):
        super().__init__(module)
        self.scopeDict = scopeDict
        self.config = None
        self.module = module
        self.module.recompile()

    @staticmethod
    def substitute_compiler_graph(
        og: "Graph",
        g: "Graph",
        val_map: Dict[Node, Node],
        target_prefix: str,
        curmod: torch.nn.Module,
        rootmod: torch.nn.Module,
        return_output_node=False,
    ) -> "Optional[Argument]":
        def my_node_copy(
            g: Graph,
            node: Node,
            curmod: torch.nn.Module,
            rootmod: torch.nn.Module,
            target_prefix: str,
            arg_transform: Callable[[Node], "Argument"] = lambda x: x,
        ) -> Node:
            """ """
            args = map_arg(node.args, arg_transform)
            kwargs = map_arg(node.kwargs, arg_transform)
            assert isinstance(args, tuple)
            assert isinstance(kwargs, dict)

            if node.op == "get_attr":
                # Fixup the target by adding in the module prefix
                curnode_target = target_prefix + "." + node.target
            else:
                curnode_target = node.target

            result_node = g.create_node(
                node.op,
                curnode_target,
                args,
                kwargs,
                node.name,
                node.type,
            )
            result_node.meta = copy.copy(node.meta)
            return result_node

        for node in g.nodes:
            if node in val_map:
                continue
            if node.op == "output":
                rv = map_arg(node.args[0], lambda n: val_map[n])
                return rv if not return_output_node else (rv, node)
            val_map[node] = my_node_copy(
                og, node, curmod, rootmod, target_prefix, lambda n: val_map[n]
            )
        return None

    def call_module(
        self, target: "Target", args: Tuple[Argument, ...], kwargs: Dict[str, Any]
    ) -> Any:
        """Check if the current module that 'target' points to is in the dmx_aware_mapping
        and substitutes in the fxir subgraph
        """
        assert isinstance(target, str)
        submod = self.fetch_attr(target)
        curr_mod = submod

        dmx_module_targets = itertools.chain(
            dmx_aware_mapping.values(), dmx_aware_functional_mappings.values()
        )
        if any(map(lambda m: isinstance(curr_mod, m), dmx_module_targets)):
            {str(v): k for k, v in dmx_aware_mapping.items()}
            subgraph = curr_mod.to_compiler_graph()
            processed_args = process_args(args)
            subgraph_input_nodes = filter(
                lambda n: n.op == "placeholder", list(subgraph.nodes)
            )
            val_map = {n: processed_args[i] for i, n in enumerate(subgraph_input_nodes)}
            curr_node = self.substitute_compiler_graph(
                self.new_graph, subgraph, val_map, target, curr_mod, self.module, False
            )
            if isinstance(curr_node, torch.fx.node.Node):
                curr_node = Proxy(curr_node)
            return curr_node
        else:
            curr_node = self.tracer.call_module(submod, submod.forward, args, kwargs)
            return curr_node

    def transform(self) -> fx.GraphModule:
        with fx_traceback.preserve_node_meta():
            result = super().run(enable_io_processing=False)
        if result is not None:

            def strip_proxy(a: Union[Argument, Proxy]) -> Any:
                return a.node if isinstance(a, Proxy) else a

            self.new_graph.output(map_aggregate(result, strip_proxy))
        return fx.GraphModule(self.module, self.new_graph)



import sys
import time
import torch
from collections import defaultdict
from dmx.compressor.modeling import DmxModel
import time
from dmx.compressor.functional import NoApproximation
from tabulate import tabulate
import enum
import gc

from typing import (
    Any,
    Dict,
    Tuple,
    Callable,
    List,
    Union,
)


class EVALUATION_MODE(enum.Enum):
    VANILLA = "Vanilla"
    BASELINE = "Baseline"
    BASIC = "Basic"
    BASIC_NOVSIMD = "Basic_NoVSIMD"


def prepare_model(
    model: torch.nn.Module,
    evaluation_mode: EVALUATION_MODE,
    model_runner: Callable[[Union[torch.nn.Module, DmxModel]], None],
):
    """Prepares a DMXModel if needed in Baseline, Basic, or Basic_NoVSIMD modes

    Parameters
    ----------
    model : torch.nn.Module
        torch model to prepare
    evaluation_mode : EVALUATION_MODE
        The mode for the DMXModel to create out of the torch model
    model_runner : Callable[[Union[torch.nn.Module, DmxModel]], None]
        A function for running a sample input through the model


    """
    vsimd_modules_by_type = defaultdict(list)

    if evaluation_mode == EVALUATION_MODE.BASELINE:
        model = DmxModel.from_torch(model)
        model.to_baseline_mode()
        model_runner(model)
    elif evaluation_mode in [EVALUATION_MODE.BASIC, EVALUATION_MODE.BASIC_NOVSIMD]:
        model = DmxModel.from_torch(model)
        model.to_basic_mode()
        model_runner(model)
        vsimd_modules = [
            (x[0], x[1])
            for x in model.named_dmx_modules()
            if not isinstance(x[1].approximator.function, NoApproximation)
        ]
        for name, m in vsimd_modules:
            if evaluation_mode == EVALUATION_MODE.BASIC_NOVSIMD:
                m.approximator.function = NoApproximation()
            vsimd_modules_by_type[type(m)].append(".".join(name.split(".")[1:]))

    return model, vsimd_modules_by_type


def measure_mode_perf(
    model: torch.nn.Module,
    model_runner: Callable[[Union[torch.nn.Module, DmxModel]], None],
    device: torch.device,
    evaluation_mode: EVALUATION_MODE,
    n_warmup_runs: int = 1,
    n_measure_runs: int = 5,
):
    """Measure model's runtime statistics for a given mode

    Parameters
    ----------
    model : torch.nn.Module
        Pytorch model to measure
    model_runner : Callable[[Union[torch.nn.Module, DmxModel]], None]
        Callable to run a sample input through the model
    device : torch.device
        device on which to run the model
    evaluation_mode : EVALUATION_MODE
        Model mode (Vanilla torch, Baseline, Basic without VSIMD ops,
        Basic)
    n_warmup_runs : int
        Number of warmup runs before gathering statistics
    n_measure_runs : int
        Number of runs across which to average the gathered statistics


    """
    assert n_warmup_runs >= 1 and n_measure_runs > 1
    model, vsimd_modules_by_type = prepare_model(model, evaluation_mode, model_runner)
    for _ in range(n_warmup_runs):
        model_runner(model)

    if evaluation_mode == EVALUATION_MODE.VANILLA:
        torch.cuda.synchronize()
        t1 = time.perf_counter()
        for _ in range(n_measure_runs):
            model_runner(model)
            torch.cuda.synchronize()
        t2 = time.perf_counter()
        return {"total_time": (t2 - t1) / n_measure_runs, "per_layer_times": {}}
    else:

        mod_names = [
            ".".join(x[0].split(".")[1:]) for x in list(model.named_dmx_modules())
        ]
        all_runtimes = []
        for _ in range(n_measure_runs):
            with model.measure_runtimes(device, mod_names):
                model_runner(model)
            all_runtimes.append(model.get_runtime_records())

        average_perlayer_runtimes = {
            k: sum([sum(single_run_runtime[k]) for single_run_runtime in all_runtimes])
            / len(all_runtimes)
            for k in mod_names
        }

        total_time = sum(average_perlayer_runtimes.values())
        return {
            "total_time": total_time,
            "per_layer_times": average_perlayer_runtimes,
            "vsimd_modules_by_type": vsimd_modules_by_type,
        }


def evaluate_vsimd_ops_deltas(results: Dict[str, Any]):
    """Generate summary table with the runtime impact of each type of vsimd op

    Parameters
    ----------
    results : Dict[str, Any]
        The runtime measuerement results

    """
    b_val = EVALUATION_MODE.BASIC.value
    b_novsimd_val = EVALUATION_MODE.BASIC_NOVSIMD.value
    if b_val in results and b_novsimd_val in results:
        vsimd_modules_by_type = results[b_novsimd_val]["vsimd_modules_by_type"]
        vsimd_time_deltas = {}
        for type_name, layer_names in vsimd_modules_by_type.items():
            time_basic = sum(
                results[b_val]["per_layer_times"][lay_name] for lay_name in layer_names
            )
            time_basic_novsimd = sum(
                results[b_novsimd_val]["per_layer_times"][lay_name]
                for lay_name in layer_names
            )
            vsimd_time_deltas[type_name] = time_basic - time_basic_novsimd

        return vsimd_time_deltas

    print(
        "Specify EVALUATION_MODE.BASIC and EVALUATION_MODE.BASIC_NOVSIMD  \
    to obtain vsimd op deltas",
        list(results.keys()),
    )
    return None


def measure_model_runtime(
    model_maker: Callable[[], Tuple[torch.nn.Module, Callable, Callable, torch.device]],
    modes: List[EVALUATION_MODE],
):
    """Entry function for measuring various runtime statistics

    Parameters
    ----------
    model_maker : Callable[[], Tuple[torch.nn.Module, Callable, Callable, torch.device]]
        A callable that returns the model to be measured, together
        with some callables to run a sample input through the model or
        to evaluate the model's accuracy
    modes : List[EVALUATION_MODE]
        List of modes on which to evaluate the model's runtime
    """
    results = {}
    layer_names = []
    for mode in modes:
        print(f"Starting runtime measurements for mode {mode.value}")
        model, model_runner, _, device = model_maker()
        torch.cuda.reset_peak_memory_stats(device)
        results[mode.value] = measure_mode_perf(model, model_runner, device, mode)
        results[mode.value]["max_memory"] = torch.cuda.max_memory_allocated()
        del model
        gc.collect()
        torch.cuda.empty_cache()
        if len(results[mode.value]["per_layer_times"]) > len(layer_names):
            layer_names = list(results[mode.value]["per_layer_times"].keys())

    main_table_str = tabulate(
        {
            " ": ["max memory (GB)", "total time (s)", *layer_names],
            **{
                k: [
                    results[k]["max_memory"] / 2**30,
                    results[k]["total_time"],
                    *results[k]["per_layer_times"].values(),
                ]
                for k in results
            },
        },
        headers="keys",
        tablefmt="github",
    )
    print(main_table_str)

    vsimd_time_deltas = evaluate_vsimd_ops_deltas(results)
    if vsimd_time_deltas is not None:
        print("*********************VSIMD operations **********\n\n\n")
        lay_types = list(vsimd_time_deltas.keys())
        lay_deltas = list(vsimd_time_deltas.values())
        cumulative_runtime = (
            (
                torch.cumsum(torch.Tensor([0] + lay_deltas), dim=0)
                + results[EVALUATION_MODE.BASIC_NOVSIMD.value]["total_time"]
            )
            .numpy()
            .tolist()
        )

        vsimd_deltas_table_str = tabulate(
            {
                "Layer type": ["(BASIC mode time without VSIMD ops)"] + lay_types,
                "Time deltas": [" "] + lay_deltas,
                "Total run time": cumulative_runtime,
            },
            headers="keys",
            tablefmt="github",
        )

        print(vsimd_deltas_table_str)


def measure_model_accuracy(
    model_maker: Callable[[], Tuple[torch.nn.Module, Callable, Callable, torch.device]],
    modes: List[EVALUATION_MODE],
):
    """Entry function for measuring a model's accuracy across various modes

    Parameters
    ----------
    model_maker : Callable[[], Tuple[torch.nn.Module, Callable, Callable, torch.device]]
        A callable that returns the model to be measured, together
        with some callables to run a sample input through the model or
        to evaluate the model's accuracy
    modes : List[EVALUATION_MODE]
        List of modes on which to measure the model's accuracy

    """
    results = {}
    for mode in modes:
        print(f"Starting evaluation for mode {mode.value}")
        model, model_runner, evaluation_fn, device = model_maker()
        model, _ = prepare_model(model, mode, model_runner)

        results[mode.value] = evaluation_fn(model, device, mode.value)

    metric_names = list(list(results.values())[0].keys())
    main_table_str = tabulate(
        {
            " ": metric_names,
            **{k: [results[k][metric] for metric in metric_names] for k in results},
        },
        headers="keys",
        tablefmt="github",
    )
    print(main_table_str)


def collect_layer_activations(
    model_maker: Callable[[], Tuple[torch.nn.Module, Callable, Callable, torch.device]],
    mode: EVALUATION_MODE,
):
    """Collect the output activations for each named DMX module in the model for the given mode

    Parameters
    ----------
    model_maker : Callable[[], Tuple[torch.nn.Module, Callable, Callable, torch.device]]
        A callable that returns the model to be measured, together
        with some callables to run a sample input through the model or
        to evaluate the model's accuracy
    mode : EVALUATION_MODE
        The mode for which we want to collect the layer activations


    """
    model, model_runner, evaluation_fn, device = model_maker()
    model, _ = prepare_model(model, mode, model_runner)

    if mode == EVALUATION_MODE.VANILLA:
        with torch.no_grad():
            final_output = model_runner(model)
        return {}, {}, final_output
    else:
        mods = [".".join(x[0].split(".")[1:]) for x in model.named_dmx_modules()]
        mods_dict = {
            ".".join(x[0].split(".")[1:]): x[1] for x in model.named_dmx_modules()
        }
        with torch.no_grad():
            with model.monitoring(mods):
                final_output = model_runner(model)

        return mods_dict, model.get_monitoring_records(), final_output


def gather_tensors(
    tensor_collection: Union[torch.Tensor, List[Any], Tuple[Any], Dict[str, Any]],
) -> List[torch.Tensor]:
    """Gathers all torch tensors from arbitrary nested structures of Lists and Dicts

    Parameters
    ----------
    tensor_collection : Union[torch.Tensor, List[Any], Tuple[Any], Dict[str, Any]]
        A Torch tensor or an arbitrary collection of tensors such as
        what you would typically get as an output from a HuggingFace
        model

    Returns
    -------
    List[torch.Tensor]



    """
    if isinstance(tensor_collection, torch.Tensor):
        return [tensor_collection]
    elif isinstance(tensor_collection, (tuple, list)):
        return [tens for x in tensor_collection for tens in gather_tensors(x)]
    elif isinstance(tensor_collection, dict):
        return [tens for v in tensor_collection.values() for tens in gather_tensors(v)]
    else:
        return []


def compute_mse_error(
    t_list1: List[torch.Tensor], t_list2: List[torch.Tensor]
) -> float:
    """Compute sum of MSE errors between corresponding pairs of tensors in the given tensor lists

    Parameters
    ----------
    t_list1 : List[torch.Tensor]
    t_list2 : List[torch.Tensor]

    Returns
    -------
    float

    """

    return sum(
        torch.nn.functional.mse_loss(x.float(), y.float()).item()
        for x, y in zip(t_list1, t_list2)
    )


def compute_maxdelta_error(
    t_list1: List[torch.Tensor], t_list2: List[torch.Tensor]
) -> float:
    """Compute the maximum delta observed between two tensor elements across all pairs of tensors in the given lists

    Parameters
    ----------
    t_list1 : List[torch.Tensor]
    t_list2 : List[torch.Tensor]

    Returns
    -------
    float


    """
    return max(
        [(x - y).float().abs().max().item() for x, y in zip(t_list1, t_list2)] + [0]
    )


def compute_error(
    out1: Union[torch.Tensor, List[Any], Tuple[Any], Dict[str, Any]],
    out2: Union[torch.Tensor, List[Any], Tuple[Any], Dict[str, Any]],
):
    """Computes the MSE error and the maximum delta between the tensors in the given pair of tensor collections

    Parameters
    ----------
    out1 : Union[torch.Tensor, List[Any], Tuple[Any], Dict[str, Any]]
    out2 : Union[torch.Tensor, List[Any], Tuple[Any], Dict[str, Any]]


    """
    t_list1 = gather_tensors(out1)
    t_list2 = gather_tensors(out2)
    return {
        "mse": compute_mse_error(t_list1, t_list2),
        "maxdelta": compute_maxdelta_error(t_list1, t_list2),
    }


def measure_model_error(
    model_maker: Callable[[], Tuple[torch.nn.Module, Callable, Callable, torch.device]],
    modes: List[EVALUATION_MODE],
    reference_mode: EVALUATION_MODE = EVALUATION_MODE.BASELINE,
):
    """Entry function for measuring the error at the output of each layer of the model for each mode relative that layer's output in the reference mode

    Parameters
    ----------
    model_maker : Callable[[], Tuple[torch.nn.Module, Callable, Callable, torch.device]]
        A callable that returns the model to be measured, togetherwith
        some callables to run a sample input through the model or to
        evaluate the model's accuracy
    modes : List[EVALUATION_MODE]
        The modes for which we want to evaluate the errors in the
        layer's activations
    reference_mode : EVALUATION_MODE
        The reference mode whose layer activations serve as the ground
        truth used to evaluate the per-layer errors for the other
        modes


    """

    print(f"collecting activations for reference {reference_mode.value}")
    _, reference_activations, reference_final_output = collect_layer_activations(
        model_maker, reference_mode
    )
    results = defaultdict(dict)
    table_lay_names = []
    for mode in modes:
        if mode == reference_mode:
            continue
        print(f"collecting activations for {mode.value}")
        mods_dict, activations, final_output = collect_layer_activations(
            model_maker, mode
        )
        if activations and reference_activations:
            lay_names = list(mods_dict.keys())
            cum_error_dict = defaultdict(lambda: defaultdict(float))
            per_layer_error_dict = defaultdict(lambda: defaultdict(float))
            input_error_dict = defaultdict(lambda: defaultdict(float))
            for l_name in lay_names:
                assert len(activations[l_name]) == len(reference_activations[l_name])
                for datum1, datum2 in zip(
                    reference_activations[l_name], activations[l_name]
                ):
                    for metric, val in compute_error(
                        datum1["input"], datum2["input"]
                    ).items():
                        input_error_dict[l_name][metric] += val

                    out1 = datum1["output"]
                    out2 = datum2["output"]
                    for metric, val in compute_error(out1, out2).items():
                        cum_error_dict[l_name][metric] += val

                    input_args = datum1["input"][0]
                    input_kwargs = datum1["input"][1]
                    model2_clean_input_output = mods_dict[l_name](
                        *input_args, **input_kwargs
                    )

                    for metric, val in compute_error(
                        out1, model2_clean_input_output
                    ).items():
                        per_layer_error_dict[l_name][metric] += val

            results[mode.value] = {
                "cumulative": cum_error_dict,
                "per_layer": per_layer_error_dict,
                "input": input_error_dict,
            }
            if len(lay_names) > len(table_lay_names):
                table_lay_names = lay_names

        results[mode.value]["final_output_error"] = compute_error(
            final_output, reference_final_output
        )

        del activations, final_output
        gc.collect()
        torch.cuda.empty_cache()

    def make_error_str(error_dict):
        return f'{error_dict["mse"]:.2g}({error_dict["maxdelta"]:.2g})'

    table_dict = {}
    error_types = ["per_layer", "cumulative", "input"]
    error_format = "mse(max delta)"
    for k in results:
        for e_type in error_types:
            if e_type in results[k]:
                table_dict[f"{k}({e_type})"] = [
                    error_format,
                    make_error_str(results[k]["final_output_error"]),
                ] + [
                    make_error_str(results[k][e_type][l_name])
                    for l_name in table_lay_names
                ]
        if not any(e_type in results[k] for e_type in error_types):
            table_dict[k] = [
                error_format,
                make_error_str(results[k]["final_output_error"]),
            ]
    table_str = tabulate(
        {
            f"error relative to {reference_mode.value}": [
                "error_format",
                "final_output_error",
            ]
            + table_lay_names,
            **table_dict,
        },
        headers="keys",
        tablefmt="github",
    )

    print(table_str)
